from core_10x.environment_variables import EnvVars

from core_10x.ts_store import TsStore, ResourceSpec
from core_10x.traitable import NamedTsStore, TsClassAssociation


def main(username: str = '', password: str = ''):
    v_main_uri: EnvVars.Var = EnvVars.var.main_ts_store_uri
    if not v_main_uri:
        print(
f"""
Environment variable {EnvVars.var_name(v_main_uri.attr_name)} is not defined
Main Traitable Store is not specified, thus you must explicitly instantiate and use a particular Traitable Store, e.g.:

    with TsStore.instance_from_uri('mongodb://public_mongo_host.abc.com/test'):     #-- no authentication
        # your code
        ...
        
    or

    #-- Authentication is required
    password = 'a VeryStrongPassword!'
    with TsStore.instance_from_uri('mongodb://protected_mongo_host.abc.com/test', username = 'xyz', password = password):
        # your code
        ...
"""
        )
        #-- there is nothing more to do
        return

    print(
f"""
Main Traitable Store is specified (Environment variable {EnvVars.var_name(v_main_uri.attr_name)} = {v_main_uri.value})
It means that Traitable objects would be sought and saved in the above store.  
"""
    )

    main_spec = TsStore.spec_from_uri(v_main_uri.value)
    is_running, with_auth = main_spec.resource_class.is_running_with_auth(main_spec.hostname())
    if not is_running:
        print(f'Main Traitable Store {v_main_uri.value} is not running')
        return

    if with_auth:
        print(f"Main Traitable Store {v_main_uri.value} requires authentication. Let's check if a Vault is specified." )

        v_vault_uri = EnvVars.var.vault_ts_store_uri
        if not v_vault_uri:
            print(f"No Vault is specified (Environment variable {EnvVars.var_name(v_vault_uri.attr_name)} is not defined). Exiting...")
            return

        print(
f"""
Vault Store is specified (Environment variable {EnvVars.var_name(v_vault_uri.attr_name)} = {v_vault_uri.value})
It means that some Traitable Stores in use (as well as other resources) require authentication.
In order to use the Vault "automagically", you need to run "add_vault utiliry" entering your password to the Vault and your XX MasterPassword.
"""
        )
        return

    main_store = main_spec.resource_class.instance(**main_spec.kwargs)
    v_class_assoc = EnvVars.var.use_ts_store_per_class
    if v_class_assoc:
        print(
f"""
Different Traitable Stores may be used for different subclasses of Traitable (Environment variable {EnvVars.var_name(v_class_assoc.attr_name)} = {v_class_assoc.value})
Checking further...
"""
        )

        main_store.begin_using()

        available_stores = NamedTsStore.load_many()


if __name__ == '__main__':

    import sys

    main(*sys.argv[1:])




from core_10x.code_samples.directories import ANIMALS, FISH


def test_flatten():
    f1 = ANIMALS.flatten()
    assert f1 == {
        ('Microorganisms',): 'Microorganisms',
        ('Microorganisms', 'Single Cell'): 'Single Cell',
        ('Microorganisms', 'Multi Cell'): 'Multi Cell',
        ('Mollusks',): 'Mollusks',
        ('Fishes',): 'Fishes',
        ('Fishes', 'Salt Water'): 'Salt Water',
        ('Fishes', 'Salt Water', 'Beluga'): 'Beluga',
        ('Fishes', 'Fresh Water'): 'Fresh Water',
        ('Amphibia',): 'Amphibia',
        ('Reptiles',): 'Reptiles',
        ('Birds',): 'Birds',
        ('Mammals',): 'Mammals',
        ('Mammals', 'Cats'): 'Cats',
        ('Mammals', 'Dogs'): 'Dogs',
        ('Mammals', 'Bears'): 'Bears',
        ('Mammals', 'Whales'): 'Whales',
        ('Mammals', 'Whales', 'Bluewhale'): 'Bluewhale',
        ('Mammals', 'Whales', 'Orca'): 'Orca',
        ('Mammals', 'Whales', 'Spermwhale'): 'Spermwhale',
        ('Mammals', 'Whales', 'Beluga'): 'Beluga',
    }

    f2 = ANIMALS.flatten(with_root=True)
    assert f2 == {('Animals',): 'Animals', **{('Animals', *k): v for k, v in f1.items()}}


def test_choices():
    c1 = FISH.choices()
    assert c1 == {
        'Pike Family': 'PkF',
        'Pike Family/Northern Pike': 'NPK',
        'Pike Family/Muskie': 'MSK',
        'Pike Family/Pickerel': 'PKL',
        'Perch Family': 'PeF',
        'Perch Family/Common Perch': 'PCH',
        'Perch Family/Yellow Perch': 'YPH',
        'Perch Family/Walleye': 'WLY',
        'Perch Family/Sagger': 'SGR',
        'Carp Family': 'CrF',
        'Carp Family/Common Carp': 'CRP',
        'Carp Family/Wild Carp': 'WCP',
        'Carp Family/Ide': 'IDE',
        'Carp Family/Bream': 'BRM',
        'Carp Family/Roach': 'RCH',
    }

    c2 = FISH.choices(with_root=True)
    assert c2 == {'Fresh Water Fish': 'FWF', **{f'Fresh Water Fish/{k}': v for k, v in c1.items()}}



from __future__ import annotations

from datetime import date, datetime

import dateutil.parser

MIN_CANONICAL_DATE = 10000101


class XDateTime:
    """
    All datetime values are in UTC time zone!
    """

    @staticmethod
    def int_to_date(v: int) -> date:
        """
        ordinal or canonical (<yyyy><mm><dd>)
        """
        if v >= MIN_CANONICAL_DATE:
            day = v % 100
            ym = v // 100

            month = ym % 100
            year = ym // 100

            return date(year=year, month=month, day=day)

        return date.fromordinal(v)

    @staticmethod
    def date_to_int(v: date, ordinal: bool = True) -> int:
        return v.toordinal() if ordinal else (10000 * v.year + 100 * v.month + v.day)

    # fmt: off
    FORMAT_X10  = '%Y%m%d'
    FORMAT_ISO  = '%Y-%m-%d'
    FORMAT_US   = '%m/%d/%Y'
    FORMAT_EU   = '%d.%m.%Y'
    # fmt: on

    s_default_format = FORMAT_X10
    formats = [s_default_format, FORMAT_X10, FORMAT_ISO, FORMAT_US, FORMAT_EU]

    @classmethod
    def set_default_format(cls, fmt: str):
        cls.s_default_format = fmt
        cls.formats[0] = fmt

    @staticmethod
    def str_to_date(v: str, format: str = '') -> date | None:
        if format:
            try:
                return datetime.strptime(v, format).date()
            except Exception:
                return None

        for fmt in XDateTime.formats:
            try:
                return datetime.strptime(v, fmt).date()
            except Exception:
                continue

        try:
            return dateutil.parser.parse(v).date()
        except Exception:
            pass

    formats_to_str = (
        f'{formats[0]} %H:%M:%S',
        f'{formats[0]} %H:%M:%S.%f',
    )

    @staticmethod
    def datetime_to_str(v: datetime, with_ms: bool = False) -> str:
        fmt = XDateTime.formats_to_str[with_ms]
        return v.strftime(fmt)

    @staticmethod
    def date_to_str(v: date, format: str = '') -> str:
        if not format:
            format = XDateTime.formats[0]
        return v.strftime(format)

    # fmt: off
    date_converters = {
        date:       lambda v:   v,
        datetime:   lambda v:   date(v.year, v.month, v.day),
        int:        int_to_date,
        str:        str_to_date,
    }
    # fmt: on
    @classmethod
    def to_date(cls, v) -> date | None:
        fn = cls.date_converters.get(type(v))
        return fn(v) if fn else None

    dt_format = ('%H:%M', '%H:%M:%S')

    @staticmethod
    def str_to_datetime(v: str) -> datetime | None:
        parts = v.split(' ')
        try:
            date_part, time_part = parts
            d = XDateTime.str_to_date(date_part)
            if d is not None:
                num_colons = time_part.count(':')
                fmt = XDateTime.dt_format[num_colons]
                if time_part.find('.') != -1:
                    fmt = fmt + '.%f'
                    return datetime.strptime(time_part, fmt)
        except Exception:
            pass

        try:
            return dateutil.parser.parse(v)
        except Exception:
            pass

    @staticmethod
    def date_to_datetime(d: date) -> datetime:
        return datetime(year=d.year, month=d.month, day=d.day)

    # fmt: off
    datetime_converters = {
        datetime:   lambda v:   v,
        date:       lambda v:   datetime(year = v.year, month = v.month, day = v.day),
        int:        lambda v:   XDateTime.date_to_datetime(XDateTime.int_to_date(v)),
        str:        str_to_datetime,
    }
    # fmt: on

    @classmethod
    def to_datetime(cls, v) -> datetime:
        fn = cls.datetime_converters.get(type(v))
        return fn(v) if fn else None



import asyncio

import rio.testing.browser_client
from ui_10x.rio.component_builder import DynamicComponent
from ui_10x.rio.widgets import Label


async def check_text(widget, test_client, text):
    assert widget['text'] == text
    client_text = await test_client.execute_js('document.querySelector(".rio-text")?.children[0]?.innerText || ""')
    assert client_text == text


async def test_label() -> None:
    """Test basic Label widget functionality with client interaction verification."""
    widget = Label('Hello World')

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test initial text - verify widget state
        assert widget['text'] == 'Hello World'

        # Test text change - verify widget update
        widget.set_text('Updated Text')
        await test_client.wait_for_refresh()
        await check_text(widget, test_client, 'Updated Text')


async def test_label_empty() -> None:
    """Test Label widget with empty text and client validation."""
    widget = Label()

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test empty text
        await check_text(widget, test_client, '')

        # Test setting text
        widget.set_text('New Text')
        await test_client.wait_for_refresh()
        await check_text(widget, test_client, 'New Text')



from __future__ import annotations

from typing import Literal

from core_10x.named_constant import Enum

import rio
import ui_10x.platform_interface as i
from ui_10x.rio.component_builder import Widget


class SCROLL(Enum):
    OFF = 'never'
    ON = 'always'
    AS_NEEDED = 'auto'


class ScrollArea(Widget, i.ScrollArea):
    s_component_class = rio.ScrollContainer

    def set_widget(self, w: Widget):
        self.set_children([w])

    def set_horizontal_scroll_bar_policy(self, h):
        self._set_scrollbar_policy('scroll_x', h)

    def set_vertical_scroll_bar_policy(self, h):
        self._set_scrollbar_policy('scroll_y', h)

    def _set_scrollbar_policy(self, scroll: Literal['scroll_x', 'scroll_y'], policy: SCROLL):
        self[scroll] = policy.label



import os

if __name__ == '__main__':
    assert os.environ.setdefault('UI_PLATFORM', 'Rio') == os.getenv('UI_PLATFORM')
    from ui_10x.platform import ux

    print(ux.Dialog(children=[ux.Label('Message'), ux.Label('Message2')], title='Title').exec())

    # from functools import partial
    #
    # from rio import project_config, App, Text, TextStyle, ComponentPage, page

    # app = App(
    #     pages=[
    #         ComponentPage( '', '',
    #             build=lambda: ux.VBoxLayout(
    #                 ux.Label('a'),
    #                 ux.PushButton('b',on_press=partial(print, 'b pressed')),
    #             )(),
    #         )
    #     ]
    # )
    # app.run_in_window()
    #



import polars as pl
from infra_10x.mongodb_store import MongoStore

from core_10x.code_samples.person import Person

if __name__=='__main__':
    with MongoStore.instance(hostname ='localhost', dbname='test', username='', password=''):
        objs = Person.load_many()
        df = pl.DataFrame([o.serialize_object() for o in objs])
        print(df)
        df.write_database('people',connection="postgresql:///postgres?host=/tmp",engine='adbc',if_table_exists='replace')

        df2 = pl.read_database('people',connection="postgresql:///postgres?host=/tmp",engine='adbc')
        print(df2)


from __future__ import annotations

from typing import TYPE_CHECKING

from core_10x.global_cache import cache
from core_10x.nucleus import Nucleus
from core_10x.ts_store import TsCollection, TsDuplicateKeyError, TsStore, standard_key
from py10x_infra import MongoCollectionHelper
from pymongo import MongoClient, errors
from pymongo.errors import DuplicateKeyError
from pymongo.uri_parser import parse_uri as pymongo_parse_uri

if TYPE_CHECKING:
    from collections.abc import Iterable

    from core_10x.ts_store import f
    from pymongo.collection import Collection
    from pymongo.database import Database


class MongoCollection(TsCollection):
    s_id_tag = '_id'

    assert Nucleus.ID_TAG() == s_id_tag, f"Nucleus.ID_TAG() must be '{s_id_tag}'"

    def __init__(self, db, collection_name: str, store: MongoStore):
        self.coll: Collection = db[collection_name]
        self.store = store

    def _session_kw(self):
        tx = self.store.current_transaction()
        return {'session': tx.session} if tx is not None else {}

    def collection_name(self) -> str:
        return self.coll.name

    def id_exists(self, id_value: str) -> bool:
        return self.coll.count_documents({self.s_id_tag: id_value}, **self._session_kw()) > 0

    def find(self, query: f = None, _at_most: int = 0, _order: dict = None) -> Iterable:
        cursor = self.coll.find(query.prefix_notation() if query else {}, **self._session_kw())
        if _order:
            cursor = cursor.sort(list(_order.items()))
        if _at_most:
            cursor = cursor.limit(_at_most)
        return cursor

    def count(self, query: f = None) -> int:
        return self.coll.count_documents(query.prefix_notation() if query else {}, **self._session_kw())

    def save_new(self, serialized_traitable: dict, overwrite: bool = False) -> int:
        needs_upsert = (set_values := serialized_traitable.get('$set')) or overwrite
        id_tag = self.s_id_tag
        id_value = (set_values or serialized_traitable)[id_tag]
        e = None

        # TODO: overwrite via save(), not save_new() so that revision is incremented rather than reset
        (set_values or serialized_traitable)[Nucleus.REVISION_TAG()] = 1

        sk = self._session_kw()
        if not needs_upsert:
            try:
                res = self.coll.insert_one(serialized_traitable, **sk)
            except DuplicateKeyError:
                ack, cnt = False, 1
            else:
                ack, cnt = res.acknowledged, 0
        else:
            res = self.coll.update_one({id_tag: id_value}, serialized_traitable if set_values else {'$set': serialized_traitable}, upsert=True, **sk)
            ack, cnt = res.acknowledged, res.matched_count

        if cnt and not overwrite:  # -- e.g. this id/revision already existed
            raise TsDuplicateKeyError(self.collection_name(), {id_tag: id_value}) from e

        return int(ack)

    def save(self, serialized_traitable: dict) -> int:
        rev_tag = Nucleus.REVISION_TAG()
        id_tag = self.s_id_tag

        revision = serialized_traitable.get(rev_tag, -1)
        assert revision >= 0, 'revision must be >= 0'

        if revision == 0:
            return self.save_new(serialized_traitable)

        id_value = serialized_traitable.get(id_tag)

        filter = {}
        pipeline = []
        serialized_traitable = dict(serialized_traitable)  # -- copy to avoid modifying the input
        MongoCollectionHelper.prepare_filter_and_pipeline(serialized_traitable, filter, pipeline)

        res = self.coll.update_one(filter, pipeline, **self._session_kw())
        if not res.acknowledged:
            return revision

        if not res.matched_count:  # -- e.g. restore from deleted
            raise AssertionError(f'{self.coll} {id_value} has been most probably inappropriately restored from deleted')

        if res.matched_count != 1:
            return revision

        return revision if res.modified_count != 1 else revision + 1

    def delete(self, id_value: str) -> bool:
        q = {self.s_id_tag: id_value}
        return self.coll.delete_one(q, **self._session_kw()).acknowledged

    def create_index(self, name: str, trait_name: str | list[tuple[str, int]], **index_args) -> str | None:
        """Create index. When inside a transaction, defers to run on commit (MongoDB disallows createIndex in txn)."""
        tx = self.store.current_transaction()
        if tx is not None:
            tx.pending_create_index.append(
                (self.collection_name(), name, trait_name, dict(index_args))
            )
            return name
        sk = self._session_kw()
        index_info = self.coll.index_information(**sk)
        if name in index_info:
            return None

        return self.coll.create_index(trait_name, name=name, **{**index_args, **sk})

    def max(self, trait_name: str, filter: f = None) -> dict | None:
        if filter:
            cur = self.coll.find(filter.prefix_notation(), **self._session_kw()).sort({trait_name: -1}).limit(1)
        else:
            cur = self.coll.find(**self._session_kw()).sort({trait_name: -1}).limit(1)
        for data in cur:
            return data

        return None

    def min(self, trait_name: str, filter: f = None) -> dict | None:
        if filter:
            cur = self.coll.find(filter.prefix_notation(), **self._session_kw()).sort({trait_name: 1}).limit(1)
        else:
            cur = self.coll.find(**self._session_kw()).sort({trait_name: 1}).limit(1)
        for data in cur:
            return data

        return None

    def load(self, id_value: str) -> dict | None:
        for data in self.coll.find({self.s_id_tag: id_value}, **self._session_kw()):
            return data

        return None




class MongoStore(TsStore, resource_name='MONGO_DB'):
    ADMIN = 'admin'

    # fmt: off
    s_instance_kwargs_map = dict(
        port    = ('port',                      27017),
        ssl     = ('ssl',                       False),
        sst     = ('serverSelectionTimeoutMS',  10000),
    )
    # fmt: on

    s_cached_connections: dict[tuple, MongoClient] = {}

    class Transaction(TsStore.Transaction):
        def __init__(self, store: MongoStore):
            if not (current_tx:=store.current_transaction()):
                self.session = session = store.client.start_session()
                session.start_transaction()
                self.pending_create_index: list[tuple[str, str, str | list[tuple[str, int]], dict]] = []
            else:
                self.session = current_tx.session
                self.pending_create_index = current_tx.pending_create_index
            super().__init__(store)

        def _do_commit(self) -> None:
            if self.store.current_transaction():
                return  # -- no nested transactions supported
            try:
                self.session.commit_transaction()
            finally:
                self.session.end_session()
            self._run_pending_create_index()

        def _run_pending_create_index(self) -> None:
            """Run create_index calls that were deferred during the transaction (MongoDB disallows createIndex in txn)."""
            for coll_name, name, trait_name, index_args in self.pending_create_index:
                coll = self.store.collection(coll_name)
                coll.create_index(name, trait_name, **index_args)
            self.pending_create_index.clear()

        def _do_abort(self) -> None:
            if self.store.current_transaction():
                return  #-- no nested transactions supported
            try:
                self.session.abort_transaction()
            finally:
                self.session.end_session()

    @classmethod
    def connect(cls, hostname: str, username: str, password: str, _cache: bool = True, _throw: bool = True, **kwargs) -> MongoClient:
        connection_key = standard_key((hostname, username), kwargs) if _cache else None
        client = cls.s_cached_connections.get(connection_key)
        if not client:
            client = MongoClient(hostname, username=username, password=password, **kwargs)
            try:
                client.server_info()
            except Exception:
                client.close()
                if _throw:
                    raise
                client = None
        if client and connection_key:
            cls.s_cached_connections[connection_key] = client

        return client

    @classmethod
    def uncache_connection(cls, hostname: str, username: str, password: str, **kwargs):
        connection_key = standard_key((hostname, username), kwargs)
        client = cls.s_cached_connections.pop(connection_key, None)
        if client:
            client.close()

    # noinspection PyMethodOverriding
    @classmethod
    def new_instance(cls, hostname: str, dbname: str, username: str, password: str, **kwargs) -> TsStore:
        client = cls.connect(hostname, username, password, **kwargs)
        return cls(client, client[dbname], username)

    @classmethod
    def parse_uri(cls, uri: str) -> dict:
        try:
            params = pymongo_parse_uri(uri)
            # fmt: off
            hostname, port  = params['nodelist'][0]
            kwargs          = params['options']
            kwargs[cls.PORT_TAG]  = port
            args = {
                cls.HOSTNAME_TAG:   hostname,
                cls.DBNAME_TAG:     params['database'],
                cls.USERNAME_TAG:   params['username'],
                cls.PASSWORD_TAG:   params['password'],
            }
            # fmt: on
            args.update(kwargs)
            return args
        except Exception as e:
            raise ValueError(f'Invalid URI = {uri}') from e

    def __init__(self, client: MongoClient, db: Database, username: str):
        super().__init__()
        self.client = client
        self.db: Database = db
        self.username = username

    def collection_names(self, regexp: str = None) -> list:
        filter = dict(name={'$regex': regexp}) if regexp else None
        return self.db.list_collection_names(filter=filter)

    def collection(self, collection_name: str) -> TsCollection:
        return MongoCollection(self.db, collection_name, store=self)

    def supports_transactions(self) -> bool:
        """True if this MongoDB deployment supports multi-document transactions (replica set or mongos)."""
        try:
            res = self.client.admin.command('ismaster')
            return 'setName' in res or res.get('msg') == 'isdbgrid'
        except Exception:
            return False

    def delete_collection(self, collection_name: str) -> bool:
        self.db.drop_collection(collection_name)
        return True

    @classmethod
    @cache
    def is_running_with_auth(cls, host_name: str) -> tuple:  # -- (is_running, with_auth)
        client = cls.connect(host_name, '', '', _cache=False, _throw=False)
        if not client:
            return False, False

        admin_db = client[cls.ADMIN]
        try:
            res = admin_db.command('getCmdLineOpts')
            auth = any(r == '--auth' for r in res['argv'][1:])
            return True, auth

        except errors.OperationFailure:  # -- auth is required
            return True, True

        finally:
            client.close()

    def auth_user(self) -> str | None:
        return self.username



from __future__ import annotations

from typing import TYPE_CHECKING

from core_10x.concrete_traits import list_trait
from core_10x.trait import Ui

from ui_10x.choice import Choice
from ui_10x.trait_widget import TraitWidget
from ui_10x.utils import UxClipBoard, ux, ux_push_button

if TYPE_CHECKING:
    from core_10x.trait import Trait


class LineEditWidget(TraitWidget, ux.LineEdit, widget_type=Ui.WIDGET_TYPE.LINE):
    def _create(self):
        ux.LineEdit.__init__(self)

        self.text_edited_connect(self.on_editing)
        self.editing_finished_connect(self.on_editing_finished)
        self.was_edited = False

    def on_editing(self, text):
        self.was_edited = True

    def on_editing_finished(self):
        if self.was_edited:
            self.was_edited = False
            str_value = self.text()
            if not str_value:  # -- user cleared the field, let's get the value back!
                self.update_trait_value(invalidate=True)
            else:
                ##value = self.trait.from_str(str_value)
                self.update_trait_value(value=str_value, invalidate=False)

    def mouse_press_event(self, event: ux.MouseEvent):
        if event.is_right_button():
            if not self.trait_editor.is_read_only():
                cb = UxClipBoard()
                cb.popup(self, self.traitable, self.trait.name)
            else:
                ux.LineEdit.mouse_press_event(self, event)

    def _set_read_only(self, flag):
        self.set_read_only(flag)

    def _value(self):
        str_value = self.text()
        return self.trait.from_str(str_value)

    def _set_value(self, value):
        str_value = self.trait.to_str(value)
        self.set_text(str_value)


def create_text_widget(editor, trait: Trait):
    if isinstance(trait, list_trait):
        return TextEditForListWidget(editor, trait)

    return TextEditWidget(editor, trait)


TraitWidget.s_hinted_widgets[Ui.WIDGET_TYPE.TEXT] = create_text_widget


class TextEditWidget(TraitWidget, ux.TextEdit):
    def _create(self):
        ux.TextEdit.__init__(self)

    def focus_out_event(self, event):
        value = self._value()
        self.update_trait_value(value=value) if value else self.update_trait_value(invalidate=True)

    def _set_read_only(self, flag):
        self.set_read_only(flag)

    def _value(self):
        return self.to_plain_text()

    def _set_value(self, value):
        if not isinstance(value, str):
            value = ''
        self.set_plain_text(value)


class TextEditForListWidget(TextEditWidget):
    def _value(self) -> list:
        text = self.to_plain_text()
        return text.split('\n')

    def _set_value(self, value):
        if value:
            if not isinstance(value, list):
                raise AssertionError('list is expected')

            value = '\n'.join(value)
            self.set_plain_text(value)


class PasswordWidget(LineEditWidget, widget_type=Ui.WIDGET_TYPE.PASSWORD):
    def _create(self):
        super()._create()
        self.set_password_mode()
        self.set_text('')


class CheckBoxWidget(TraitWidget, ux.CheckBox, widget_type=Ui.WIDGET_TYPE.CHECK):
    def _create(self):
        ux.CheckBox.__init__(self)
        if self.trait.ui_hint.param('right_label', False):
            self.set_text(self.trait.ui_hint.label)

        self.state_changed_connect(self.on_state_changed)

    def on_state_changed(self, flag):
        self.update_trait_value(value=bool(flag))

    def _set_read_only(self, flag):
        self.set_enabled(flag)

    def _value(self):
        return bool(self.is_checked())

    def _set_value(self, value):
        self.set_checked(bool(value))


class ChoiceWidget(TraitWidget, ux.Widget, widget_type=Ui.WIDGET_TYPE.CHOICE):
    def _create(self):
        ux.Widget.__init__(self)

        lay = ux.HBoxLayout()
        lay.set_spacing(0)
        lay.set_contents_margins(0, 0, 0, 0)
        self.set_layout(lay)

        # -- Line edit
        self.line_edit = le = ux.LineEdit()
        if self.trait_editor.is_read_only():
            le.set_read_only(True)

        le.editing_finished_connect(self.on_editing_finished)
        lay.add_widget(le)

        # -- Arrow down button
        self.button = ux_push_button('', callback=self.show_popup, style_icon='ArrowDown')
        lay.add_widget(self.button)

        traitable = self.trait_editor.traitable
        trait = self.trait
        self.choice = Choice(f_choices=lambda: traitable.get_choices(trait), f_selection_callback=lambda item: self.on_selection(item))

    def _value(self):
        selection = self.choice.values_selected
        return selection[0] if selection else None

    def _set_value(self, value):
        if not self.choice.directory:
            value = self.choice.inverted_choices.get(value)

        if not value:
            value = ''
        self.line_edit.set_text(value)

    def _set_read_only(self, flag):
        self.line_edit.set_read_only(flag)
        self.button.set_enabled(not flag)

    # def on_current_index_changed(self, index):
    #     self.update_trait_value()

    def show_popup(self):
        self.popup = d = self.choice.popup(parent=self, _open=False)
        w = self.width()
        h = self.height()
        xy = self.map_to_global(ux.Point(0, 0))

        d.set_modal(True)
        d.show()
        d.set_geometry(xy.x(), xy.y(), max(w, d.width()), h)

    def on_selection(self, item):
        self.line_edit.set_text(item)
        self.update_trait_value(value=self.choice.values_selected[0], invalidate=False)

    def on_editing_finished(self):
        str_value = self.line_edit.text()
        if not str_value:
            self.update_trait_value(invalidate=True)
        else:
            value = self.choice.choices.get(str_value, str_value)
            self.update_trait_value(value=value, invalidate=False)


class PixmapLabelWidget(TraitWidget, ux.Label, widget_type=Ui.WIDGET_TYPE.PIXMAP):
    def _create(self):
        ux.Label.__init__(self)
        # self.pixmap = UxPixmap()

    ...


class PushButtonWidget(TraitWidget, ux.PushButton, widget_type=Ui.WIDGET_TYPE.PUSH):
    def _create(self):
        ux.PushButton.__init__(self)
        self.set_text(self.trait.ui_hint.label)

        callback_name = f'{self.trait.name}_action'
        ed = self.trait_editor
        traitable = ed.traitable.traitable
        cb = getattr(traitable.__class__, callback_name, None)
        if cb:
            self.clicked_connect(lambda dummy: cb(traitable, ed))

    ...

    def _set_read_only(self, flag):
        pass

    def _value(self):
        pass

    def _set_value(self, value):
        if value is None:
            value = True
        self.set_enabled(value)



from __future__ import annotations

import bisect
import math
from datetime import date, timedelta
from typing import Any

from scipy import interpolate

from core_10x.named_constant import NamedConstant
from core_10x.traitable import RC, RC_TRUE, RT, AnonymousTraitable, M, T, Traitable


class IP_KIND(NamedConstant, lowercase_values = True):
    #-- NOTE: labels specify the minimum number of points required!
    LINEAR      = (2, )
    NEAREST     = (2, )
    NEAREST_UP  = (2, )
    PREVIOUS    = (2, )     #-- return the previous value of the point
    NEXT        = (2, )     #-- ------ the next ----
    ZERO        = (1, )     #-- spline interp of zeroth order
    SLINEAR     = (2, )     #-- spline interp of first order
    QUADRATIC   = (3, )     #-- spline interp of second order
    CUBIC       = (4, )     #-- spline interp of third order

    NO_INTERP   = (0, )     #-- NO interp outside of given nodes

class CurveParams(Traitable):
    DEFAULT_INTERPOLATOR = interpolate.interp1d

    interpolator: Any   = RT(default = DEFAULT_INTERPOLATOR)
    ip_kind: IP_KIND    = RT(IP_KIND.LINEAR)
    assume_sorted: bool = RT(True)
    copy: bool          = RT(False)
    fill_value: Any     = RT('extrapolate')     ## it's 'extrapolate' or a tuple (left_value, right_value) for extrapolation
    bounds_error: bool  = RT(False)

class Curve(AnonymousTraitable):
    times: list         = T([])       #-- only ints or floats are allowed
    values: list        = T([])
    params: CurveParams = RT()

    beginning_of_time: Any  = T(None)    #-- may be float or int

    interpolator: Any       = RT()
    min_curve_size: int     = RT()

    def params_get(self) -> CurveParams:
        return CurveParams()

    def min_curve_size_get(self) -> int:
        return self.params.ip_kind.label

    def start_time(self):
        times = self.times
        return times[0] if times else None

    def end_time(self):
        times = self.times
        return times[-1] if times else None

    def update(self, t, value, reset=True):
        if type(value) is not float:  # -- TODO: we sometimes have np.floats
            value = float(value)

        times = self.times
        values = self.values
        i = bisect.bisect_left(times, t)
        if i >= len(times):
            times.append(t)
            values.append(value)
        else:
            if times[i] == t:
                values[i] = value
            else:
                times.insert(i, t)
                values.insert(i, value)

        self.times = times
        self.values = values
        if reset:
            self.reset()

    def update_many(self, times, values, reset=True):
        assert len(times) == len(values), 'times and values size mismatch'
        for i, t in enumerate(times):
            self.update(t, values[i], reset=False)

        if reset:
            self.reset()

    def remove(self, t, reset=True) -> bool:
        times = self.times
        if t not in times:
            return False

        values = self.values
        i = times.index(t)
        times.pop(i)
        values.pop(i)
        if reset:
            self.reset()

        return True

    def set_curve_params(self, **param_values) -> RC:
        rc = self.params.set_values(**param_values)
        self.reset()
        return rc

    def set_curve_params_to_extrapolate(self) -> RC:
        return self.set_curve_params(bounds_error = False, fill_value = 'extrapolate')

    ## extrapolate flat left (by the first value) and right (by last value)
    def set_curve_params_to_flat_extrapolate(self) -> RC:
        values = self.values
        return self.set_curve_params(bounds_error = False, fill_value = (values[0], values[-1]))

    def interpolator_get(self):
        params = self.params
        return params.interpolator(
            self.times, self.values,
            kind            = params.ip_kind.value,
            assume_sorted   = params.assume_sorted,
            copy            = params.copy,
            fill_value      = params.fill_value,
            bounds_error    = params.bounds_error,
        )

    def value(self, t) -> float:
        times = self.times

        if self.params.ip_kind is IP_KIND.NO_INTERP:
            return self.values[times.index(t)] if t in times else math.nan

        if len(times) < self.min_curve_size:
            if t in times:
                return self.values[times.index(t)]

        bot = self.beginning_of_time
        return float(self.interpolator(t)) if (bot is not None) or (t >= bot) else math.nan

    def values_at(self, dates) -> tuple:
        return tuple(self.value(d) for d in dates)

    def reset(self):
        self.invalidate_value('interpolator')


class TwoFuncInterpolator:
    def __init__(self, in_func, out_func, in_func_on_arrays=None, _interpolator=interpolate.interp1d):
        if in_func:
            in_func_on_arrays = lambda list_x, list_y: [in_func(x, list_y[i]) for i, x in enumerate(list_x)]

        assert in_func_on_arrays, 'either in_func or in_func_on_arrays must be provided'

        self.in_func = in_func_on_arrays
        self.out_func = out_func
        self.interpolator = _interpolator

    def __call__(self, x, y, **kwargs):
        values = self.in_func(x, y)
        ip_method = self.interpolator(x, values, **kwargs)
        out_func = self.out_func

        def _method(t):
            v = ip_method(t)
            return out_func(t, v)

        return _method


class DateCurve(Curve):
    beginning_of_time: int  = M()
    dates: list             = RT()

    s_epoch_date = date(1970, 1, 1)

    @classmethod
    def _to_number(cls, d) -> int:
        if isinstance(d, date):
            return (d - cls.s_epoch_date).days
        elif isinstance(d, int):
            return d

        raise ValueError(f'Unexpected value {d}')

    @classmethod
    def _from_number(cls, x: int) -> date:
        return cls.s_epoch_date + timedelta(days=x)

    def dates_get(self) -> list:
        f = self._from_number
        return [f(x) for x in self.times]

    def dates_set(self, trait, value) -> RC:
        f = self._to_number     #-- TODO: possibly improve performance by using a different f (which doesn't check the type)
        times = [f(d) for d in value]
        return self.set_value('times', times)

    def beginning_of_time_set(self, trait, d) -> RC:
        t = self._to_number(d)
        self.raw_set_value(trait, t)
        return RC_TRUE

    def beginning_of_time_as_date(self) -> date:
        return self._from_number(self.beginning_of_time)

    def start_time(self) -> date:
        times = self.dates
        return times[0] if times else None

    def end_time(self) -> date:
        times = self.dates
        return times[-1] if times else None

    def update(self, d: date, value, reset=True):
        x = self._to_number(d)
        super().update(x, value, reset=reset)

    def reset(self):
        super().reset()
        self.invalidate_value('dates')

    def value(self, d: date) -> float:
        x = self._to_number(d)
        return super().value(x)

    def remove(self, d, reset=True):
        t = self._to_number(d)
        if super().remove(t, reset=reset):
            self.invalidate_value('dates')

    def dates_values(self, min_date: date = None, max_date: date = None) -> list:
        dates = self.dates
        values = self.values
        return [(d, v) for d, v in zip(dates, values, strict=True) if (not min_date or d >= min_date) and (not max_date or d <= max_date)]



from __future__ import annotations

import collections
import contextlib
import re
import uuid
from collections import Counter
from contextlib import nullcontext
from datetime import date
from typing import TYPE_CHECKING, Any
from typing_extensions import Self

import numpy as np
import pytest
from core_10x import trait_definition
from core_10x.code_samples.person import Person
from core_10x.exec_control import BTP, CACHE_ONLY, GRAPH_ON, INTERACTIVE
from core_10x.py_class import PyClass
from core_10x.rc import RC, RC_TRUE
from core_10x.trait import Trait
from core_10x.trait_definition import RT, M, T, TraitDefinition, TraitModification
from core_10x.trait_method_error import TraitMethodError
from core_10x.traitable import THIS_CLASS, AnonymousTraitable, Traitable, TraitAccessor
from core_10x.traitable_id import ID
from core_10x.ts_store import TsStore
from core_10x.xnone import XNone

if TYPE_CHECKING:
    from collections.abc import Generator


class SubTraitable(Traitable):
    trait1: int = RT(T.ID)
    trait2: str = RT()

    x = '123'  # -- not in slots


class SubTraitable2(SubTraitable):
    trait1 = M(flags=(None, T.ID))  # removes ID flag
    trait2: int = M()
    trait3: float = T() // 'trait definition comment'
    trait4: int = RT(0)


class SubTraitable3(SubTraitable):
    trait2: list[str] = M() // 'trait modification comment'


def test_subclass_traits():
    expected_traits = {'trait1', 'trait2'}
    assert {t.name for t in SubTraitable.traits(flags_off=T.RESERVED)} == expected_traits


def test_subclass2_traits():
    expected_traits = ['trait1', 'trait2', 'trait3', 'trait4']
    assert [t.name for t in SubTraitable2.traits(flags_off=T.RESERVED)] == expected_traits
    assert SubTraitable.trait('trait2').data_type is str
    assert SubTraitable2.trait('trait2').data_type is int
    assert SubTraitable3.trait('trait2').data_type is list
    assert SubTraitable2.trait('trait4').default_value() == 0
    assert SubTraitable2.trait('trait2').ui_hint.tip == 'Trait2'
    assert SubTraitable2.trait('trait3').ui_hint.tip == 'trait definition comment'
    assert SubTraitable3.trait('trait2').ui_hint.tip == 'trait modification comment'


def test_is_storable():
    assert not SubTraitable.is_storable()
    assert SubTraitable2.is_storable()

    with pytest.raises(RuntimeError, match='No Traitable Store is specified: neither explicitly, nor via environment variable XX_MAIN_TS_STORE_URI'):
        SubTraitable2().save().throw()

    assert 'is not storable' in SubTraitable(trait1=uuid.uuid1().int).save().error()


def test_trait_update():
    with CACHE_ONLY():
        instance = SubTraitable(trait1=10, trait2='hello', _replace=True)
        assert instance.trait2 == 'hello'

        assert instance == SubTraitable.update(trait1=10, trait2='world')
        assert instance.trait2 == 'world'

        assert instance == SubTraitable.update(trait1=10, trait2=None)  # setting to None
        assert instance.trait2 is None


def test_instance_slots():
    assert Traitable.__slots__ == ()
    assert SubTraitable.__slots__ == ()
    instance = SubTraitable(trait1=10)
    assert not hasattr(instance, '__dict__')
    with pytest.raises(AttributeError):
        instance.non_existent_attr = 'value'


def test_post_init_called():
    calls = []

    class X(Traitable):
        x: int = RT()

        def __post_init__(self):
            calls.append(self)

    x = X(x=1)
    assert calls == [x]


def test_overriding_init_disallowed():
    with pytest.raises(
        TypeError,
        match=r'Overriding __init__ is not allowed in DisallowedInit\. Use __post_init__ instead\.',
    ):

        class DisallowedInit(Traitable):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)


def test_init_with_id():
    with GRAPH_ON():  # isolate lazy reference so it does not affect other tests
        pid = ID('John|Smith')
        p = Person(pid)
        assert p.id() == pid


def test_init_with_trait_values():
    with CACHE_ONLY():
        p = Person(first_name='John', last_name='Smith')
    assert p.first_name == 'John'
    assert p.last_name == 'Smith'


def test_set_values():
    with CACHE_ONLY():
        p = Person(first_name='John', last_name='Smith')
    rc = p.set_values(age=19, weight_lb=200)
    assert rc


def test_dynamic_traits():
    class X(Traitable):
        s_own_trait_definitions = dict(x=RT(data_type=int, default=10))

    x = X()
    assert x.T.x.data_type is int
    assert x.x == 10

    class Y(X):
        y: int = RT(20)

    y = Y()
    assert y.T.x.data_type is int
    assert y.x == 10

    assert y.T.y.data_type is int
    assert y.y == 20


def test_collection_name_trait():
    class X(Traitable):
        x: int

    assert not X.is_storable()
    assert not X.trait('_collection_name')

    class Y(Traitable):
        s_default_trait_factory = T
        s_custom_collection = True
        y: int

        @classmethod
        def load_data(cls, id: ID) -> dict | None:
            return None

    assert Y.is_storable()
    assert Y.trait('_collection_name')

    y = Y(_collection_name='test')

    assert y.id().collection_name == 'test'
    assert y._collection_name == 'test'


@pytest.mark.parametrize('on_graph', [0, 1])
@pytest.mark.parametrize('debug', [0, 1])
@pytest.mark.parametrize('convert_values', [0, 1])
@pytest.mark.parametrize('use_parent_cache', [True, False])
@pytest.mark.parametrize('use_default_cache', [True, False])
@pytest.mark.parametrize('use_existing_instance_by_id', [True, False])
@pytest.mark.parametrize('self_ref', [True, False])
@pytest.mark.parametrize('nested', [True, False])
def test_traitable_ref_load(on_graph, debug, convert_values, use_parent_cache, use_default_cache, use_existing_instance_by_id, self_ref, nested):
    load_calls = collections.Counter()

    class X(Traitable):
        i: int = T(T.ID)
        x: THIS_CLASS = T()
        y: int = T()

        @classmethod
        def exists_in_store(cls, id: ID) -> bool:
            return id.value == '1'

        @classmethod
        def load_data(cls, id: ID) -> dict | None:
            v = id.value
            load_calls[v] += 1
            i = int(v)
            return {'_id': v, 'i': i, '_rev': 1, 'y': 1} | ({'x': {'_id': str(i + int(not self_ref))}} if i < 3 else {})

    with BTP.create(on_graph, convert_values, debug, use_parent_cache, use_default_cache):
        x = X.existing_instance_by_id(ID('1')) if use_existing_instance_by_id else X.existing_instance(i=1)
        assert x
        x1 = X(i=3, x=x, y=2, _replace=True)
        assert x1.x is x
        assert x1.y == 2  # still 2 as lazy-load occurs before setting parameters passed
        assert load_calls == {'3': 1}  # lazy-load since no db access in constructor
        load_calls.clear()

        assert x.i == 1
        expected = lambda n: {str(i): 1 for i in range(1, n + 1)}
        if debug and not self_ref:
            assert load_calls == expected(3)
            assert x.x.x == x1  # found existing instance
            assert x1.x is XNone  # reload in debug mode
        else:
            with BTP.create(-1, -1, -1, use_parent_cache=False, use_default_cache=False) if nested else nullcontext():
                assert load_calls == expected(1)
                assert x.x
                assert not self_ref or x == x.x
                assert load_calls == expected(1)
                assert x.x.i == 1 + int(not self_ref)
                assert load_calls == expected(1 + int(not self_ref))
                assert x.x.x == (x if self_ref else x1)  # found existing instance
                assert x1.x is x  # no reload

    # TODO: change_flags; as_of context
    # TODO: nodes with args...


def test_trait_methods():
    class A(Traitable):
        s_default_trait_factory = T
        t: int

        @classmethod
        def exists_in_store(cls, id):
            return False

        @classmethod
        def load_data(cls, id):
            return None

    class B(A):
        def t_get(self):
            return 1

        @classmethod
        def t_serialize(cls, trait, value):
            return value + 1

    class C(B):
        t: str = M()

        def t_get(self):
            return 2

        @classmethod
        def t_serialize(cls, trait, value):
            return value + 2

    class D(C):
        t: date = M()

    for t, dt in zip([A, B, C, D], [int, int, str, date], strict=True):
        assert t.trait('t').data_type is dt

    for t, v in zip([A, B, C, D], [XNone, 1, 2, 2], strict=True):
        assert t().t is v
        assert t().serialize_object()['t'] == (v * 2 or None)


def test_serialize_rt_object():
    class X(Traitable):
        x: int = RT(T.ID)
        y: int = RT(T.ID)

    assert not X.existing_instance_by_id(ID('1|2'), _throw=False)  # not found, can't find by id
    x = X.existing_instance(x=1, y=2)  # creates new instance by id traits
    assert x == X.existing_instance_by_id(ID('1|2'))
    assert x == X(x=1, y=2)

    s = x.serialize(False)
    assert s == {'_id': [1, 2]}

    assert X.deserialize(s) == x


def test_trait_modification_inheritance_with_flags():
    """Test complex trait modification inheritance with M() overriding flags and getters."""

    class X(Traitable):
        x: int = RT(T.ID)
        v: int = RT(T.ID)

        def v_get(self):
            return self.x + 1

    class Y(X):
        # M() removes ID flag; getter from X is inherited
        v: int = M(flags=(None, T.ID), default=3)

    class Z(Y):
        # M() changes data type; value is provided by v_get defined here
        v: float = M(default=XNone)

        def v_get(self):
            return float(self.x + 3)

    x = X(x=1)
    y = Y(x=1)
    z = Z(x=1)

    # X: v is computed via v_get
    assert x.v == 2  # x + 1

    # Y: v uses default
    assert y.v == 3

    # Z: v uses its own v_get and returns float
    assert z.v == 4.0  # x + 3, as float

    # Verify trait definitions
    assert X.trait('v').flags_on(T.ID)
    assert not Y.trait('v').flags_on(T.ID)  # ID flag removed
    assert Z.trait('v').data_type is float

    assert X.trait('v').default is XNone
    assert Y.trait('v').default == 3
    assert Z.trait('v').default is XNone


def test_anonymous_traitable(monkeypatch):
    monkeypatch.setattr('core_10x.package_refactoring.PackageRefactoring.default_class_id', lambda cls, *args, **kwargs: PyClass.name(cls))

    class X(AnonymousTraitable):
        a: int = T()

    class Y(Traitable):
        y: int = T(T.ID)
        x: Traitable = T()

        @classmethod
        def exists_in_store(cls, id):
            return False

        @classmethod
        def load_data(cls, id):
            return None

    class Z(Y):
        x: AnonymousTraitable = M(T.EMBEDDED)

    with pytest.raises(RuntimeError, match=r"T1.x - class <class 'core_10x.traitable.Traitable'> must be declared 'embeddable'"):
        class _T1(Traitable):
            x: Traitable = T(T.EMBEDDED)

    class _T2(Traitable):
        x: AnonymousTraitable = T()

    x = X(a=1)
    assert x.serialize(True) == {'a': 1}

    y = Y(y=0, x=x, _replace=True)
    s = y.serialize_object()
    assert s['x'] == {'_obj': {'a': 1}, '_type': '_nx', '_cls': 'test_traitable.test_anonymous_traitable.<locals>.X'}

    z = Z(y=1, x=x, _replace=True)
    s = z.serialize_object()
    assert s['x'] == {'_obj': {'a': 1}, '_type': '_nx', '_cls': 'test_traitable.test_anonymous_traitable.<locals>.X'}

    z = Z(y=2, x=Y(y=3), _replace=True)
    with pytest.raises(TraitMethodError, match=r"test_anonymous_traitable.<locals>.Y/3 - embedded instance must be 'embeddable'"):
        z.serialize_object()


class TestSerializeEmbedded:
    class Y(Traitable, embeddable=True):
        y: int = T()
        a: int = T()

    class Z(Traitable):
        z: int = T(T.ID)
        a: int = T()

    class X1(Traitable):
        x: int = T(T.ID)
        a: AnonymousTraitable = T(T.EMBEDDED)
        b: list[Traitable] = T()

    class X2(X1):
        b: list[Traitable] = M(T.EMBEDDED)

    @pytest.mark.parametrize('cls', [X1, X2])
    def test_serialize_embedded(self,cls):
        Y, Z = self.Y, self.Z  # noqa: N806
        with CACHE_ONLY():
            x = cls(
                x=1,
                a=Y(y=1, a=1, _replace=True),
                b=[
                    Y(y=2, a=2, _replace=True),
                    Z(z=3, a=4, _replace=True),
                ],
                _replace=True,
            )
            try:
                serialized = x.serialize_object()
                assert serialized == {
                    '_id': '1',
                    '_rev': 0,
                    'x': 1,
                    'a': {'_type': '_nx', '_cls': 'test_traitable/TestSerializeEmbedded/Y', '_obj': {'y': 1, 'a': 1}},
                    'b': [
                        None,
                        {'_type': '_nx', '_cls': 'test_traitable/TestSerializeEmbedded/Y', '_obj': {'y': 2, 'a': 2}},
                        {'_type': '_nx', '_cls': 'test_traitable/TestSerializeEmbedded/Z', '_obj': {'_id': '3'}},
                    ],
                }
                z = cls.deserialize(serialized).b[1]
                assert isinstance(z, Z)
                assert z.a == 4
            except TraitMethodError as e:
                if cls is self.X2:
                    assert "embedded instance must be 'embeddable'" in str(e)
                else:
                    raise




def test_own_trait_defs():
    cnt = Counter()

    def assert_and_cont(what, obj, arg, trait=None):
        assert isinstance(obj, Traitable)
        if trait:
            assert isinstance(trait, Trait)
            assert trait.data_type is int
            assert trait.flags_on(T.RUNTIME | T.EXPENSIVE)
        cnt[what] += arg

    class X(Traitable):
        def x_get(self, arg) -> int:
            assert_and_cont('get', self, arg)
            return 1

        def x_set(self, trait, value, arg) -> RC:
            assert_and_cont('set', self, arg, trait)
            self.raw_set_value(trait, value + 1, arg)
            return RC_TRUE

        @classmethod
        def own_trait_definitions(cls) -> Generator[tuple[str, trait_definition.TraitDefinition]]:
            yield 'x', TraitDefinition(T.RUNTIME | T.EXPENSIVE, data_type=int)

    t = X.trait('x')
    assert cnt == {}
    x = X()
    assert x.x(1) == 1
    assert cnt == {'get': 1}

    x.set_value(t, 2, 1)
    assert x.x(1) == 3
    assert cnt == {'get': 1, 'set': 1}


def test_trait_get_default_override():
    class X(Traitable):
        x: int = RT(default=1)

    assert X().x == 1

    with pytest.raises(
        RuntimeError,
        match=r'Ambiguous definition for x_get on <class \'test_traitable.test_trait_get_default_override.<locals>.Y\'> - both x.default and <class \'test_traitable.test_trait_get_default_override.<locals>.Y\'>.x_get are defined.',
    ):

        class Y(X):
            x: int = RT(default=2)

            def x_get(self):
                return 2

    class Z(X):
        def x_get(self):
            return 3

    class T(Z):
        def x_get(self):
            return 4

    class S(T):
        x: int = RT(default=5)

    class R(S):
        x: int = RT(default=XNone)

    with pytest.raises(
        RuntimeError,
        match=r'Ambiguous definition for x_get on <class \'test_traitable.test_trait_get_default_override.<locals>.P\'> - both x.default and <class \'test_traitable.test_trait_get_default_override.<locals>.P\'>.x_get are defined.',
    ):

        class P(S):
            x: int = M()

            def x_get(self):
                return 6

    assert Z().x == 3
    assert T().x == 4
    assert S().x == 5
    assert R().x == 4


def test_create_and_share():
    class X(Traitable):
        x: int = RT(T.ID)
        y: int = RT(T.ID)
        z: int = RT()
        t: int = RT(T.ID_LIKE)

        def y_get(self):
            return self.t

    #   with pytest.raises(TypeError, match=re.escape('test_create_and_share.<locals>.X expects at least one ID trait value')):
    with pytest.raises(TypeError, match=re.escape("test_create_and_share.<locals>.X.x (<class 'int'>) - invalid value ''")):
        X()

    with pytest.raises(TypeError, match=re.escape("test_create_and_share.<locals>.X.y (<class 'int'>) - invalid value ''")):
        X(x=1)

    X(x=1, y=1, z=1, _replace=True)

    with pytest.raises(ValueError, match=re.escape('test_create_and_share.<locals>.X.z - non-ID trait value cannot be set during initialization')):
        X(x=1, y=1, z=2)

    with pytest.raises(ValueError, match=re.escape('test_create_and_share.<locals>.X.z - non-ID trait value cannot be set during initialization')):
        X(x=1, y=1, z=1)

    with pytest.raises(ValueError, match=re.escape('test_create_and_share.<locals>.X.z - non-ID trait value cannot be set during initialization')):
        X(x=1, t=1, z=1)

    assert X(x=1, t=1).z == 1
    assert X(x=1, y=1).z == 1

    with INTERACTIVE():
        x = X()  # empty object allowed - OK!
        z = X(x=1, y=1)
        assert z.z == 1  # found from parent

        x.x = 1
        x.y = 1
        assert not x.share(False)
        assert x.z is XNone


def test_serialize(monkeypatch):
    monkeypatch.setattr('core_10x.package_refactoring.PackageRefactoring.default_class_id', lambda cls, *args, **kwargs: PyClass.name(cls))
    save_calls = Counter()
    history_save_calls = Counter()
    load_calls = Counter()
    serialized = {}

    class X(Traitable):
        x: int = T(T.ID)
        y: Self = T()
        z: int = T()

        @classmethod
        def exists_in_store(cls, id: ID) -> bool:
            return id.value in serialized or int(id.value) > 3

        @classmethod
        def load_data(cls, id: ID) -> dict | None:
            load_calls[id.value] += 1
            return serialized.get(id.value)

        @classmethod
        def store(cls):
            class Store:
                def auth_user(self):
                    return 'test_user'

                def transaction(self):
                    return contextlib.nullcontext()

                def populate(self, params, serialized_data):
                    return TsStore.populate(self, params, serialized_data)

                def collection(self, collection_name):
                    class Collection:
                        def create_index(self, name, trait_name):
                            return name

                        def save(self, serialized_data):
                            if not collection_name.endswith('#history'):
                                id_value = serialized_data['_id']
                                save_calls[id_value] += 1
                            else:
                                id_value = serialized_data['$set']['_id']
                                history_save_calls[id_value] += 1

                            serialized[id_value] = serialized_data
                            return 1

                        save_new = save

                    return Collection()

            return Store()

        def z_get(self) -> int:
            return self.y._rev if self.y and self.y._rev else 0

    class Y(X): ...

    x = X(x=0)
    assert not serialized
    x.save().throw()
    assert not load_calls
    assert dict(save_calls) == {'0': 1}
    assert dict(history_save_calls) == {next(iter(history_save_calls)): 1}
    assert serialized['0']['z'] == 0
    save_calls.clear()
    load_calls.clear()

    x = X(x=1, y=X(x=2, y=X(x=1), _replace=True), _replace=True)
    x.save()
    assert save_calls == {'1': 1}
    assert load_calls == {str(i): 1 for i in range(1, 3)}
    save_calls.clear()
    load_calls.clear()

    x.save(save_references=True)
    assert save_calls == {'1': 1, '2': 1}
    assert load_calls == {}
    save_calls.clear()

    X(x=3, y=Y(_id=ID('4')), z=0, _replace=True).save(save_references=True)
    assert load_calls == {'3': 1}
    assert save_calls == {'3': 1}  # save of a lazy load is noop

    assert X(x=3).y.__class__ is Y
    with pytest.raises(
        RuntimeError,
        match=r"Failed in <class 'test_traitable.test_serialize.<locals>.X'>.z.z_get\n    object = 5;\noriginal exception = RuntimeError: test_serialize.<locals>.X/6: object reference not found in store",
    ):
        X(x=5, y=X(_id=ID('6')), _replace=True).save(save_references=True).throw()


def test_reference_serialization_roundtrip(monkeypatch):
    """Test that references are correctly serialized and can be loaded back in a round-trip."""
    monkeypatch.setattr('core_10x.package_refactoring.PackageRefactoring.default_class_id', lambda cls, *args, **kwargs: PyClass.name(cls))
    serialized = {}

    class Person(Traitable):
        first_name: str = T(T.ID)
        last_name: str = T(T.ID)
        spouse: Self = T()

        @classmethod
        def exists_in_store(cls, id: ID) -> bool:
            return id.value in serialized

        @classmethod
        def load_data(cls, id: ID) -> dict | None:
            return serialized.get(id.value)

        @classmethod
        def store(cls):
            class Store:
                def auth_user(self):
                    return 'test_user'

                def transaction(self):
                    return contextlib.nullcontext()

                def collection(self, collection_name):
                    class Collection:
                        def create_index(self, name, trait_name):
                            return name

                        def save(self, serialized_data):
                            id_value = serialized_data['_id']
                            serialized[id_value] = serialized_data
                            return 1

                        save_new = save

                    return Collection()

            return Store()

    # Create both people
    p2 = Person(first_name='Tatiana', last_name='Pevzner', _replace=True)
    p1 = Person(first_name='Ilya', last_name='Pevzner', spouse=p2, _replace=True)

    # Manually serialize both with references (simulating save_references=True behavior)
    serialized['Tatiana|Pevzner'] = p2.serialize_object()
    serialized['Ilya|Pevzner'] = p1.serialize_object()
    assert serialized['Ilya|Pevzner']['spouse'] == {'_id': 'Tatiana|Pevzner'}

    # Reload and verify references are preserved
    loaded_p1 = Person.load(ID('Ilya|Pevzner'))
    assert loaded_p1.first_name == 'Ilya'
    assert loaded_p1.spouse is not None
    assert loaded_p1.spouse.first_name == 'Tatiana'
    assert loaded_p1.spouse.last_name == 'Pevzner'
    assert type(loaded_p1.spouse) is Person


def test_id_trait_set():
    class X(Traitable):
        x: int = RT(T.ID)
        y: int = RT(T.ID_LIKE)

    with INTERACTIVE():
        x = X()
        x.x = 1
        x.y = 1
        assert x.x == 1
        x.x = 2
        assert x.x == 2

        x.share(False)
        with pytest.raises(ValueError, match=r"test_id_trait_set.<locals>.X.x \(<class 'int'>\) - cannot change ID trait value from '2' to '3'"):
            x.x = 3

        with pytest.raises(ValueError, match=r"test_id_trait_set.<locals>.X.x \(<class 'int'>\) - cannot change ID trait value from '2' to '2'"):
            x.x = 2
        assert x.x == 2

        with pytest.raises(ValueError, match=r"test_id_trait_set.<locals>.X.y \(<class 'int'>\) - cannot change ID_LIKE trait value from '1' to '4'"):
            x.y = 4

        with pytest.raises(ValueError, match=r"test_id_trait_set.<locals>.X.y \(<class 'int'>\) - cannot change ID_LIKE trait value from '1' to '1'"):
            x.y = 1

        assert x.y == 1

    x = X(x=1)
    with pytest.raises(ValueError, match=r"test_id_trait_set.<locals>.X.x \(<class 'int'>\) - cannot change ID trait value from '1' to '2'"):
        x.x = 2

    with pytest.raises(ValueError, match=r"test_id_trait_set.<locals>.X.x \(<class 'int'>\) - cannot change ID trait value from '1' to '1'"):
        x.x = 1
    assert x.x == 1


def test_reload():
    rev = 0

    class X(Traitable):
        x: int = T(T.ID)

        @classmethod
        def load_data(cls, id: ID) -> dict | None:
            nonlocal rev
            rev += 1
            data = {'_id': id.value, 'x': int(id.value), '_rev': rev}
            return data

    # reload of lazy ref
    x = X(ID('1'))
    x.reload()
    assert x._rev == 1

    x.reload()
    assert x._rev == 2

    x = X(ID('2'))
    with GRAPH_ON():
        x.reload()
        assert x._rev == 3
        assert rev == 3
    assert x._rev == 3
    assert rev == 3

    x = X(ID('3'))
    with GRAPH_ON():
        y = X(ID('3'))
        assert y._rev == 4
    assert x._rev == 4


def test_separation():
    class X(Traitable):
        x: int = RT(T.ID)
        y: int = RT()

    with GRAPH_ON() as g1:
        x1 = X(x=1, y=1, _replace=True)

    with GRAPH_ON() as g2:
        x2 = X(x=1, y=2, _replace=True)

    assert X(x=1).y is XNone

    with g1:
        assert X(x=1).y == 1
        with pytest.raises(RuntimeError, match=r'X/1: object not usable - origin cache is not reachable'):
            x2.get_value('y')

    with g2:
        assert X(x=1).y == 2
        with pytest.raises(RuntimeError, match=r'X/1: object not usable - origin cache is not reachable'):
            x1.get_value('y')

    with pytest.raises(RuntimeError, match=r'X/1: object not usable - origin cache is not reachable'):
        x1.get_value('y')

    with pytest.raises(RuntimeError, match=r'X/1: object not usable - origin cache is not reachable'):
        x2.get_value('y')


def test_id_like_with_replace_and_non_id_update():
    class Cross(Traitable):
        cross: str = RT(T.ID)  # e.g., GBP/USD or CHF/JPY
        base_ccy: str = RT(T.ID_LIKE)  # Base (left) currency
        quote_ccy: str = RT(T.ID_LIKE)  # Quote (right) currency
        value: str = RT()
        value2: str = RT()

        def cross_get(self) -> str:
            return f'{self.base_ccy}/{self.quote_ccy}'

    c1 = Cross(base_ccy='A', quote_ccy='B', value='123', value2='456', _replace=True)
    assert c1.cross == 'A/B'
    assert c1.id().value == 'A/B'
    assert c1.value == '123'
    assert c1.value2 == '456'

    c2 = Cross(base_ccy='A', quote_ccy='B')
    assert c2.cross == 'A/B'
    assert c2.id().value == 'A/B'
    assert c2.value == '123'
    assert c2.value2 == '456'

    c3 = Cross(base_ccy='A', quote_ccy='B', value='234', _replace=True)
    assert c2.cross == 'A/B'
    assert c2.id().value == 'A/B'
    assert c2.value == '234'
    assert c2.value2 is XNone

    assert c1 == c2 == c3  # all objects with the same ID are equal

    with pytest.raises(
        ValueError,
        match=re.escape('test_id_like_with_replace_and_non_id_update.<locals>.Cross.value - non-ID trait value cannot be set during initialization'),
    ):
        Cross(base_ccy='A', quote_ccy='B', value='234')


class TestExistingInstance:
    class TestablePerson(Person):
        @classmethod
        def exists_in_store(cls, id: ID) -> bool:
            return id.value == 'Ariella|Pevzner'

        @classmethod
        def load_data(cls, id: ID) -> dict | None:
            return {'_id': id.value, '_rev': 1} if id.value == 'Ariella|Pevzner' else None

    class RuntimePerson(Person):
        @classmethod
        def own_trait_definitions(cls) -> Generator[tuple[str, TraitDefinition]]:
            for trait_name, trait in Person.s_dir.items():
                if not trait.flags_on(T.RUNTIME | T.RESERVED):
                    yield trait_name, TraitModification(flags=T.RUNTIME).apply(trait.t_def)

    def test_existing_instance_api_variants(self):
        assert not self.RuntimePerson.is_storable()

        if 1:
            id_traits = [
                dict(last_name='Davidovich', first_name='Sasha'),
                dict(last_name='Pevzner', first_name='Ilya'),
                dict(last_name='Lesin', first_name='Alex'),
                dict(last_name='Smith', first_name='John'),
            ]
            kwargs = [
                dict(weight_lbs=170, _replace=True),
                dict(),
                dict(weight_lbs=190, _replace=True),
                dict(),
            ]
            classes = [
                self.TestablePerson,
                self.TestablePerson,
                self.RuntimePerson,
                self.RuntimePerson,
            ]
            people = [cls(**id_traits, **kwargs) for cls, id_traits, kwargs in zip(classes, id_traits, kwargs, strict=True)]

            ppl_found = [obj.__class__.existing_instance(**id_values) for obj, id_values in zip(people, id_traits, strict=True)]
            assert ppl_found == people

            # Positive / negative lookups with _throw flag
            assert self.RuntimePerson.existing_instance(last_name='Smith', first_name='John')

            assert not self.TestablePerson.existing_instance(last_name='Pevzner', first_name='Arthur', _throw=False)
            assert not self.TestablePerson.existing_instance_by_id(ID('Arthur|Pevzner'), _throw=False)

            assert self.TestablePerson.existing_instance(last_name='Pevzner', first_name='Ariella', _throw=False)
            assert self.TestablePerson.existing_instance_by_id(ID('Ariella|Pevzner'))

            for i, cls in enumerate(people):
                # existing_instance_by_id with ID and raw id value
                id1 = ppl_found[i].id()
                p3 = cls.existing_instance_by_id(_id=id1)

                id1_val = id1.value
                p4 = cls.existing_instance_by_id(_id_value=id1_val)
                assert p3 == p4

                id2_val = 'Smith|Josh'
                p5 = cls.existing_instance_by_id(_id_value=id2_val, _throw=False)
                assert p5 is None


@pytest.mark.parametrize('value', [{'x': 1}, {}, [], [1], np.float64(1.1)])
def test_any_trait(value):
    class X(Traitable):
        x: int = T(T.ID)
        y: Any = T()
        z: list = T()

        @classmethod
        def load_data(cls, id: ID) -> dict | None:
            return None

    with GRAPH_ON():
        x = X(x=1, y=value, z=[value], _replace=True)
        assert x.y == value
        assert x.z[0] == value
        assert type(x.y) is type(value)
        assert type(x.z[0]) is type(value)
        s = x.serialize_object()

    with GRAPH_ON():
        x = X.deserialize_object(x.s_bclass, None, s)
        assert x.y == value
        assert x.z[0] == value
        assert type(x.y) is type(value)
        assert type(x.z[0]) is type(value)
        assert s == x.serialize_object()


@pytest.mark.parametrize('t', [T, RT])
def test_exceptions(t):
    class X(Traitable):
        x: int = t(T.ID)
        y: int = t(default=1)

    with CACHE_ONLY():
        x = X(_id=ID('1'))
        match = (
            r'test_exceptions.<locals>.X/1: object is an invalid lazy reference to non-storable that does not exist in memory'
            if t is RT
            else r'test_exceptions.<locals>.X/1: object reference not found in store'
        )
        with pytest.raises(RuntimeError, match=match):
            assert x.y == 1


def test_multiple_inheritance():
    """Test universal multiple inheritance with __slots__=() and MRO resolution"""

    # Test 1: Basic multiple inheritance works
    class EntityA(Traitable):
        name: str = RT(T.ID)
        value: int = RT()

        def value_get(self) -> int:
            return 42

    class EntityB(Traitable):
        rev: int = RT(T.ID)
        value: str = RT()

        def value_get(self) -> str:
            return 'forty two'

    class Combined(EntityA, EntityB):
        extra: str = RT()

    assert EntityA.__slots__ == ()
    assert EntityB.__slots__ == ()
    assert Combined.__slots__ == ()

    # Verify MRO
    mro_names = [cls.__name__ for cls in Combined.__mro__]
    assert mro_names[:4] == ['Combined', 'EntityA', 'EntityB', 'Traitable']

    obj = Combined(name='test', rev=1)

    assert obj.name == 'test'
    assert obj.value == 42
    assert obj.rev == 1

    assert isinstance(obj.T, TraitAccessor)
    assert obj.T.name.data_type is str
    assert obj.T.value.data_type is int
    assert obj.T.rev.data_type is int


def test_mutable_default():
    class X(Traitable):
        x: dict = RT(default={})

    x = X()
    y = X()
    with GRAPH_ON():
        x.x.update(a=1)

        assert x.x == {'a': 1}
        assert y.x == {}


def test_runtime_id_error():
    with pytest.raises(
        RuntimeError,
        match=r"<class 'test_traitable.test_runtime_id_error.<locals>.X'>.x is a RUNTIME ID trait - traitable must not be storable \(all traits must be RUNTIME\)",
    ):

        class X(Traitable):
            x: int = RT(T.ID)
            y: int = T()


def test_serialize_runtime_exogenous_reference():
    class X(Traitable):
        x: int = RT()

    x = X(x=1)
    with pytest.raises(
        TypeError,
        match=rf'test_serialize_runtime_exogenous_reference.<locals>.X/{x.id_value()} - cannot serialize a non-storable exogenous reference',
    ):
        x.serialize(False)


def test_serialize_runtime_endogenous_reference():
    class X(Traitable):
        x: int = RT(T.ID)

    ser = X(x=1).serialize_nx(False)
    assert ser == {'_id': [1]}

    x = Traitable.deserialize_nx(X.s_bclass, ser)
    assert isinstance(x, X)
    assert x.x == 1

    class Y(Traitable):
        x: int = RT()

    with pytest.raises(TypeError, match=r'Expected a string ID value, but found a list'):
        Y.deserialize_nx(Y.s_bclass, ser)


def test_verify_success():
    """verify() returns success when all verifiers pass and NOT_EMPTY constraints are satisfied."""

    class WithVerifier(Traitable):
        code: str = RT()
        limit: int = RT(0)

        def code_verify(self, t, value: str) -> RC:
            if value and not value.isalnum():
                return RC(False, 'code must be alphanumeric')
            return RC_TRUE

        def limit_verify(self, t, value: int) -> RC:
            if value is not None and value > 100:
                return RC(False, 'limit must be <= 100')
            return RC_TRUE

    e = WithVerifier(code='valid', limit=50)
    rc = e.verify()
    assert rc
    rc.throw()


def test_verify_fails_when_verifier_returns_error():
    """verify() returns failure and error message when a verifier fails."""

    class WithVerifier(Traitable):
        code: str = RT()

        def code_verify(self, t, value: str) -> RC:
            if value and not value.isalnum():
                return RC(False, 'code must be alphanumeric')
            return RC_TRUE

    e = WithVerifier()
    e.code = 'invalid-code'
    assert e.code == 'invalid-code'
    rc = e.verify()
    assert not rc
    assert 'code' in rc.error().lower() or 'alphanumeric' in rc.error().lower()
    assert 'Failed in' in rc.error() or 'verify' in rc.error().lower()


def test_verify_not_called_on_set():
    """Verifiers are not run on assignment; only on verify() or save(). Invalid value can be set."""

    class WithVerifier(Traitable):
        limit: int = RT(0)

        def limit_verify(self, t, value: int) -> RC:
            if value is not None and value > 100:
                return RC(False, 'limit must be <= 100')
            return RC_TRUE

    e = WithVerifier(limit=50)
    e.limit = 150  # setter not defined; verifier not run on set
    assert e.limit == 150
    rc = e.verify()
    assert not rc
    assert '100' in rc.error()



import os
import sys
import typing

if typing.TYPE_CHECKING:
    import platform_interface as ux
else:
    pname = os.getenv('UI_PLATFORM')
    if pname is None:
        pname = 'Qt6' if 'rio' not in sys.modules else 'Rio'

    if pname == 'Qt6':
        import ui_10x.qt6.platform_implementation as ux
    elif pname == 'Rio':
        import ui_10x.rio.platform_implementation as ux
    else:
        raise ImportError(f'UI implementation for `{pname}` is not available')


__all__ = ['ux']



from core_10x.rc import RC, RC_TRUE
from core_10x.trait import Trait
from core_10x.trait_definition import RT, M, T, Ui
from core_10x.traitable import Traitable


# -- A higher-level Traitable to be used for general purposes
class Entity(Traitable): ...



from __future__ import annotations

import abc
from collections import deque
from contextlib import ExitStack, contextmanager
from typing import TYPE_CHECKING

from core_10x.environment_variables import EnvVars
from core_10x.exec_control import ProcessContext
from core_10x.global_cache import standard_key
from core_10x.py_class import PyClass
from core_10x.rc import RC
from core_10x.resource import TS_STORE, Resource, ResourceSpec
from core_10x.trait_filter import f
from core_10x.ts_store_type import TS_STORE_TYPE
from py10x_kernel import BTraitableProcessorSetValueTracker as BTPTracker

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from core_10x.traitable import Traitable


class TsDuplicateKeyError(Exception):
    """Raised when attempting to insert a document with a duplicate key."""

    def __init__(self, collection_name: str, duplicate_key: dict):
        super().__init__(f'Duplicate key error collection {collection_name} dup key: {duplicate_key} was found while insert was attempted.')


class TsCollection(abc.ABC):
    s_id_tag: str = None

    @abc.abstractmethod
    def collection_name(self) -> str: ...
    @abc.abstractmethod
    def id_exists(self, id_value: str) -> bool: ...
    @abc.abstractmethod
    def find(self, query: f = None, _at_most: int = 0, _order: dict = None) -> Iterable: ...
    @abc.abstractmethod
    def count(self, query: f = None) -> int: ...
    @abc.abstractmethod
    def save_new(self, serialized_traitable: dict, overwrite: bool = False) -> int: ...
    @abc.abstractmethod
    def save(self, serialized_traitable: dict) -> int: ...
    @abc.abstractmethod
    def delete(self, id_value: str) -> bool: ...
    @abc.abstractmethod
    def create_index(self, name: str, trait_name: str | list[tuple[str, int]], **index_args) -> str: ...
    @abc.abstractmethod
    def max(self, trait_name: str, filter: f = None) -> dict: ...
    @abc.abstractmethod
    def min(self, trait_name: str, filter: f = None) -> dict: ...

    def exists(self, query: f) -> bool:
        return self.count(query) > 0

    def load(self, id_value: str) -> dict | None:
        for data in self.find(f(**{self.s_id_tag: id_value})):
            return data

    def copy_to(self, to_coll: TsCollection, overwrite: bool = False) -> RC:
        """Copy all documents from this collection to another collection."""
        rc = RC(True)

        for doc in self.find():
            try:
                if not to_coll.save_new(doc, overwrite=overwrite):
                    rc.add_error(f'Failed to save {doc.get(to_coll.s_id_tag)} to {to_coll.collection_name()}')
            except TsDuplicateKeyError:
                if overwrite:
                    raise  # -- we do not expect an exception in case of overwrite, so raise

        return rc



class TsStore(Resource, resource_type=TS_STORE):

    class Transaction:
        ended: bool = False

        def __init__(self, store: TsStore):
            self.store = store
            store.begin_transaction(self)

        def commit(self) -> None:
            """Commit the transaction. No-op if already ended."""
            if self.ended:
                return
            self.store.end_transaction(self)
            self._do_commit()

        def abort(self) -> None:
            """Abort the transaction. No-op if already ended."""
            if self.ended:
                return
            self.store.end_transaction(self)
            self._do_abort()


        def _do_commit(self) -> None: ...

        def _do_abort(self) -> None: ...

    def __init__(self):
        self._active_transactions = deque()

    def current_transaction(self) -> Transaction|None:
        return self._active_transactions[-1] if self._active_transactions else None

    def begin_transaction(self, transaction: Transaction) -> None:
        self._active_transactions.append(transaction)

    def end_transaction(self, transaction: Transaction) -> None:
        if transaction != self.current_transaction():
            raise RuntimeError(f'Transaction {transaction} is not currently active.')
        if transaction:
            self._active_transactions.pop()
            transaction.ended = True

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        assert cls.__mro__[1] is TsStore, 'TsStore must be the first base class'
        cls.s_instance_kwargs_map = {**TsStore.s_instance_kwargs_map, **cls.s_instance_kwargs_map}

    @staticmethod
    def store_class(store_class_name: str):
        cls = PyClass.find(store_class_name, TsStore)
        assert cls, f'Unknown TsStore class {store_class_name}'
        return cls

    @classmethod
    def standard_key(cls, *args, **kwargs) -> tuple:
        return standard_key(args, kwargs)

    s_instances = {}

    @classmethod
    def spec_from_uri(cls, uri: str) -> ResourceSpec:
        parts = uri.split(':', maxsplit=1)
        protocol = parts[0]
        ts_class = TS_STORE_TYPE.ts_store_class(protocol)
        return ResourceSpec(ts_class, ts_class.parse_uri(uri))

    @classmethod
    def instance(cls, *args, password: str = '', _cache: bool = True, **kwargs) -> TsStore:
        translated_kwargs = cls.translate_kwargs(kwargs)
        try:
            if not _cache:
                return cls.new_instance(*args, password=password, **translated_kwargs)

            instance_key = cls.standard_key(*args, **kwargs)
            store = cls.s_instances.get(instance_key)
            if not store:
                store = cls.new_instance(*args, password=password, **translated_kwargs)
                cls.s_instances[instance_key] = store

            return store

        except Exception as e:
            raise OSError(f'Failed to connect to {cls.s_driver_name}({args}, {translated_kwargs})\nOriginal Exception:\n{e!s}') from e

    # fmt: off
    s_instance_kwargs_map = {
        Resource.HOSTNAME_TAG:  (Resource.HOSTNAME_TAG, None),
        Resource.USERNAME_TAG:  (Resource.USERNAME_TAG, None),
        Resource.DBNAME_TAG:    (Resource.DBNAME_TAG,   None),
        Resource.PORT_TAG:      (Resource.PORT_TAG,     None),
        Resource.SSL_TAG:       (Resource.SSL_TAG,      True),
        'sst':                  ('sst',                 1000),
    }
    # fmt: on

    @classmethod
    def translate_kwargs(cls, kwargs: dict) -> dict:
        kwargs_map = cls.s_instance_kwargs_map
        def_kwargs = {name: def_value for name, (real_name, def_value) in kwargs_map.items()}
        def_kwargs.update(kwargs)
        return {kwargs_map[name][0]: value for name, value in def_kwargs.items()}

    @classmethod
    def new_instance(cls, *args, password: str, **kwargs) -> TsStore:
        raise NotImplementedError

    @classmethod
    def is_running_with_auth(cls, host_name: str) -> tuple:   # -- (is_running, with_auth)
        raise NotImplementedError

    def on_enter(self):
        self.bpc_flags = ProcessContext.reset_flags(ProcessContext.CACHE_ONLY)

    def on_exit(self):
        ProcessContext.replace_flags(self.bpc_flags)

    @abc.abstractmethod
    def collection_names(self, regexp: str = None) -> list: ...

    @abc.abstractmethod
    def collection(self, collection_name: str) -> TsCollection: ...

    @abc.abstractmethod
    def delete_collection(self, collection_name: str) -> bool: ...

    @abc.abstractmethod
    def auth_user(self) -> str | None: ...

    def supports_transactions(self) -> bool:
        return True

    def populate(self, params: list[str], serialized_data: dict):
        """Populate specified params on the server, if possible.
        This should be overridden in subclasses as needed"""
        populated_data = {
            '$set': serialized_data,
        }
        for param in sorted(params, reverse=True):
            if param == '_who':
                serialized_data['_who'] = self.auth_user()
                continue
            if param == '_at':
                del serialized_data['_at']
                populated_data['$currentDate'] = {'_at': True}
                continue
            raise KeyError(param)
        return populated_data

    def copy_to(self, to_store: TsStore, overwrite: bool = False) -> RC:
        """Copy all collections from this store to another store."""
        rc = RC(True)

        for collection_name in self.collection_names():
            from_coll = self.collection(collection_name)
            to_coll = to_store.collection(collection_name)
            rc += from_coll.copy_to(to_coll, overwrite=overwrite)

        return rc

    @contextmanager
    def transaction(self):
        """Context manager for transactional operations. Yields a transaction object with commit() and abort()."""
        tx = self.Transaction(self)
        success = False
        try:
            yield tx
            success = True
        finally:
            if success:
                tx.commit()
            else:
                tx.abort()


@contextmanager
def SaveIfChanged(classes: Sequence[type[Traitable]] = ()):  # noqa: N802
    if any(not cls.is_storable() for cls in classes):
        raise RuntimeError('Classes passed to SaveIfChanged must be storable.')

    if not isinstance(classes,tuple):
        classes = tuple(classes)

    tracker = BTPTracker()
    tracker.begin_using()
    yield tracker
    tracker.end_using()

    tracked = tuple(traitable for traitable in tracker.tracked_objects() if traitable.is_storable())

    with ExitStack() as tx_stack:
        if EnvVars.use_ts_store_transactions:
            for store in {cls.store() for cls in classes or tuple({traitable.__class__ for traitable in tracked})}:
                tx_stack.enter_context(store.transaction())

        for traitable in tracked:
            if not classes or isinstance(traitable, classes):
                traitable.save().throw()



from __future__ import annotations
import sys
from typing import Callable, Any

from PyQt6.QtWidgets import QWidget, QLayout, QBoxLayout, QVBoxLayout, QHBoxLayout, QFormLayout, QApplication
from PyQt6.QtWidgets import QLabel, QCalendarWidget, QMessageBox, QGroupBox, QButtonGroup, QRadioButton
from PyQt6.QtWidgets import QListWidgetItem, QListWidget, QLineEdit, QPlainTextEdit, QTreeWidgetItem, QTreeWidget, QCheckBox, QComboBox
from PyQt6.QtWidgets import QFrame, QSplitter, QDialog, QScrollArea, QPushButton, QStyle, QSizePolicy, QStyleOptionHeader
from PyQt6.QtWidgets import QHeaderView, QTableView
from PyQt6.QtCore import Qt, QObject, QThread, pyqtSignal, QBuffer, QIODevice, QPoint, QModelIndex    #, QByteArray
from PyQt6.QtGui import QColor, QGuiApplication, QPixmap, QMouseEvent, QFontMetrics, QPalette, QPainter
from PyQt6.QtGui import QStandardItemModel, QStandardItem

import platform

from core_10x.global_cache import cache

def missing_attr(self, item):
    t = item.title()
    item = t[0].lower() + t[1:].replace('_', '')
    return QObject.__getattribute__(self, item)

Object = QObject

QueuedConnection                = Qt.ConnectionType.QueuedConnection
AutoConnection                  = Qt.ConnectionType.AutoConnection
DirectConnection                = Qt.ConnectionType.DirectConnection
UniqueConnection                = Qt.ConnectionType.UniqueConnection
BlockingQueuedConnection        = Qt.ConnectionType.BlockingQueuedConnection

signal_decl = pyqtSignal

MouseEvent                      = QMouseEvent
MouseEvent.is_left_button       = lambda self: self.button() == Qt.MouseButton.LeftButton
MouseEvent.is_right_button      = lambda self: self.button() == Qt.MouseButton.RightButton

Point                           = QPoint

SizePolicy                      = QSizePolicy
SizePolicy.MINIMUM_EXPANDING    = QSizePolicy.Policy.MinimumExpanding
SizePolicy.PREFERRED            = QSizePolicy.Policy.Preferred

StandardItemModel               = QStandardItemModel
StandardItemModel.__getattr__   = missing_attr

StandardItem                    = QStandardItem
ModelIndex                      = QModelIndex

HeaderView                      = QHeaderView
HeaderView.section_resized_connect  = lambda self, bound_method: self.sectionResized.connect(bound_method)

Color                           = QColor

FontMetrics                     = QFontMetrics
FontMetrics.__getattr__         = missing_attr

Widget                          = QWidget
Widget.__getattr__              = missing_attr

Horizontal                      = Qt.Orientation.Horizontal
Vertical                        = Qt.Orientation.Vertical

Layout                          = QLayout

BoxLayout                       = QBoxLayout
BoxLayout.__getattr__           = missing_attr

HBoxLayout                      = QHBoxLayout
VBoxLayout                      = QVBoxLayout

FormLayout                      = QFormLayout
FormLayout.__getattr__          = missing_attr

Label                           = QLabel
Label.set_text                  = QLabel.setText

Splitter                        = QSplitter

Style                           = QStyle
Style.State_Active              = QStyle.StateFlag.State_Active

Style.__getattr__               = missing_attr

PushButton                      = QPushButton
PushButton.clicked_connect      = lambda self, bound_method:    self.clicked.connect(bound_method)

LineEdit                        = QLineEdit
LineEdit.text_edited_connect    = lambda self, bound_method:    self.textEdited.connect(bound_method)
LineEdit.editing_finished_connect = lambda self, bound_method:  self.editingFinished.connect(bound_method)
LineEdit.set_password_mode      = lambda self:                  self.setEchoMode(QLineEdit.EchoMode.Password)

TextEdit                        = QPlainTextEdit

CheckBox                        = QCheckBox
CheckBox.state_changed_connect  = lambda self, bound_method:    self.stateChanged.connect(bound_method)

ComboBox                        = QComboBox

GroupBox                        = QGroupBox

RadioButton                     = QRadioButton

ButtonGroup                     = QButtonGroup
ButtonGroup.__getattr__         = missing_attr

ListItem                        = QListWidgetItem
ListItem.__getattr__            = missing_attr

MatchExactly                    = Qt.MatchFlag.MatchExactly

ListWidget                      = QListWidget
ListWidget.clicked_connect      = lambda self, bound_method:    self.clicked.connect(bound_method)

TreeItem                        = QTreeWidgetItem
TreeItem.__getattr__            = missing_attr

TreeWidget                      = QTreeWidget
TreeWidget.item_clicked_connect = lambda self, bound_method:    self.itemClicked.connect(bound_method)
TreeWidget.item_expanded_connect = lambda self, bound_method:   self.itemExpanded.connect(bound_method)
TreeWidget.item_pressed_connect = lambda self, bound_method:    self.itemPressed.connect(bound_method)
TreeWidget.item_changed_connect = lambda self, bound_method:    self.itemChanged.connect(bound_method)

CalendarWidget                  = QCalendarWidget
CalendarWidget.selected_date    = lambda self:                  self.selectedDate().toPyDate()

class Dialog(QDialog):
    s_open_dialogs = set()
    def __init__(self,parent=None,*args,**kwargs):
        super().__init__(parent or init().activeModalWidget(),*args,**kwargs)

    def _on_finished(self):
        open_dialogs = self.s_open_dialogs
        in_open_dialogs = self in open_dialogs
        print(f'finishing {self}: {in_open_dialogs}')

        if in_open_dialogs:
            self.s_open_dialogs.remove(self)

    def show(self):
        print(f'showing {self}')
        self.s_open_dialogs.add(self)
        self.finished.connect(self._on_finished)
        super().show()
        self.raise_()
        self.activateWindow()
        self.s_open_dialogs.add(self)


class MessageBox(QMessageBox):
    s_boxes_by_parent = {}
    def __init__(self, on_close: Callable[[Any],None] = None, parent: Widget = None):
        super().__init__(parent)
        self._on_close = on_close
        self.finished.connect(self._handle_finished)

        # Set the default values for the message box
        self.setWindowFlags(Qt.WindowType.Dialog | Qt.WindowType.WindowCloseButtonHint|Qt.WindowType.WindowStaysOnTopHint)
        self.setStandardButtons(QMessageBox.StandardButton.Ok)
        self.setDefaultButton(QMessageBox.StandardButton.Ok)
        self.setEscapeButton(QMessageBox.StandardButton.Ok)

        self.setModal(True)
        self.s_boxes_by_parent[self.parent()] = self

    def _handle_finished(self, result):
        if self._on_close:
            self._on_close(result)
        self.s_boxes_by_parent.pop(self.parent())
        self.deleteLater()

    @classmethod
    def _create_message_box(cls, parent: Widget, title, message, on_close: Callable[[Any],None], icon) -> MessageBox:
        msg_box = cls(on_close,parent)
        msg_box.setWindowTitle(title)
        msg_box.setText(message)
        msg_box.setIcon(icon)
        return msg_box

    @classmethod
    def question(cls, parent: Widget, title: str, message: str, on_close: Callable[[Any],None]) -> MessageBox.StandardButton:
        msg_box = cls._create_message_box(
            parent=parent,
            title=title,
            message=message,
            on_close=on_close,
            icon=QMessageBox.Icon.Question,
        )

        msg_box.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
        msg_box.setDefaultButton(QMessageBox.StandardButton.No)
        msg_box.setEscapeButton(QMessageBox.StandardButton.No)

        if on_close:
            msg_box.show()
        else:
            return msg_box.exec()

    @classmethod
    def warning(cls, parent: Widget, title: str, message: str, on_close: Callable[[Any],None]) -> MessageBox.StandardButton:
        msg_box = cls._create_message_box(
                parent=parent,
                title=title,
                message=message,
                on_close=on_close,
                icon=QMessageBox.Icon.Warning,
          )

        if on_close:
            msg_box.show()
        else:
            return msg_box.exec()

    @classmethod
    def information(cls, parent: Widget, title: str, message: str, on_close: Callable[[Any],None]) -> MessageBox.StandardButton:
        msg_box = cls._create_message_box(
                parent=parent,
                title=title,
                message=message,
                on_close=on_close,
                icon=QMessageBox.Icon.Information,
          )

        if on_close:
            msg_box.show()
        else:
            return msg_box.exec()

    @classmethod
    def is_yes_button(cls, sb: QMessageBox.StandardButton) -> bool:
        return sb == QMessageBox.StandardButton.Yes


Application                     = QApplication

class TEXT_ALIGN:
    TOP       = Qt.AlignmentFlag.AlignTop
    V_CENTER  = Qt.AlignmentFlag.AlignVCenter
    BOTTOM    = Qt.AlignmentFlag.AlignBottom
    LEFT      = Qt.AlignmentFlag.AlignLeft
    CENTER    = Qt.AlignmentFlag.AlignCenter
    RIGHT     = Qt.AlignmentFlag.AlignRight

class SCROLL:
    OFF         = Qt.ScrollBarPolicy.ScrollBarAlwaysOff
    ON          = Qt.ScrollBarPolicy.ScrollBarAlwaysOn
    AS_NEEDED   = Qt.ScrollBarPolicy.ScrollBarAsNeeded

ScrollArea                      = QScrollArea

ForegroundRole                  = Qt.ItemDataRole.ForegroundRole
BackroundRole                   = Qt.ItemDataRole.BackgroundRole

Palette                         = QPalette
Palette.ButtonText              = QPalette.ColorRole.ButtonText
Palette.Button                  = QPalette.ColorRole.Button
Palette.Window                  = QPalette.ColorRole.Window

StyleOptionHeader               = QStyleOptionHeader
Separator                       = QLabel

@cache
def init(style = '') -> QApplication:
    if not style:
        style = platform.system()

    app = QApplication([])
    app.setStyle(style)

    if sys.platform == 'darwin':
        from ui_10x.qt6.mac_os_global_commit_filter import MacEditCommitFilter
        MacEditCommitFilter.install(app)

    return app

def to_clipboard(text: str, **kwargs):
    cp = QGuiApplication.clipboard()
    cp.setText(text)

def from_clipboard(**kwargs) -> str:
    cp = QGuiApplication.clipboard()
    return cp.text()

def separator(horizontal = True) -> Separator:
    direction = QFrame.Shape.HLine if horizontal else QFrame.Shape.VLine
    sep = QLabel()
    sep.setFrameStyle(direction | QFrame.Shadow.Sunken)
    return sep

def is_ui_thread() -> bool:
    return QThread.currentThread() is QGuiApplication.instance().thread()





from PyQt6.QtCore import QObject, QEvent
from PyQt6.QtGui import QGuiApplication
from PyQt6.QtWidgets import QApplication, QLineEdit, QAbstractSpinBox


class MacEditCommitFilter(QObject):
    def __init__(self):
        super().__init__()
        self._in_commit = False

    def eventFilter(self, obj, event):
        if self._in_commit:
            return False

        if event.type() == QEvent.Type.MouseButtonPress:
            self._commit_if_click_outside_focused_editor(event)

        return False

    def _commit_if_click_outside_focused_editor(self, event):
        focused = QApplication.focusWidget()
        if focused is None:
            return

        global_pos = event.globalPosition().toPoint()
        clicked = QApplication.widgetAt(global_pos)

        #-- If clicking inside the same widget, do nothing
        w = clicked
        while w is not None:
            if w is focused:
                return
            w = w.parentWidget()

        #-- Commit before click is delivered
        self._in_commit = True
        try:
            QGuiApplication.inputMethod().commit()

            if isinstance(focused, QLineEdit):
                focused.clearFocus()

            elif isinstance(focused, QAbstractSpinBox):
                focused.interpretText()
                focused.clearFocus()

            else:
                commit = getattr(focused, 'commitEditor', None)
                if callable(commit):
                    commit()

        finally:
            self._in_commit = False

    F = None
    @classmethod
    def install(cls, app: QApplication):
        if cls.F is None:
            f = cls()
            app.installEventFilter(f)
            cls.F = f



from datetime import datetime

from core_10x.xnone import XNone
from core_10x.resource import Resource
from core_10x.attic.vault.vault_traitable import VaultTraitable, T, RT
from core_10x.attic.vault.vault_user import VaultUser

class ResourceAccessor(VaultTraitable):
    # fmt: off
    username: str           = T(T.ID)
    resource_uri: str       = T(T.ID)

    login: str              = T()
    password: bytes         = T()

    last_updated: datetime  = T(T.EVAL_ONCE)

    user: VaultUser         = RT(T.EVAL_ONCE)
    resource: Resource      = RT(T.EVAL_ONCE)
    # fmt: on

    def username_get(self) -> str:
        return VaultUser.myname()

    def last_updated_get(self) -> datetime:
        return datetime.utcnow()

    def user_get(self) -> VaultUser:
        return VaultUser.existing_instance(user_id = self.username)

    def resource_get(self) -> Resource:
        return Resource.instance_from_uri(
            self.resource_uri,
            username = self.username,
            password = self.user.sec_keys.decrypt_text(self.password)
        )

class Vault:
    @classmethod
    def save_resource_accessor(cls, resource_uri: str, password: str, login: str = None, username: str = XNone):
        if login is None:
            login = username

        ra = ResourceAccessor(username = username, resource_uri = resource_uri)
        user = ra.user
        ra.set_values(
            login = login,
            password = user.sec_keys.encrypt_text(password)
        ).throw()

        ra.save().throw()

    #-- strictly speaking this method isn't necessary as it just returning existing_instance()
    @classmethod
    def retrieve_resource_accessor(cls, resource_uri: str, username: str = XNone) -> ResourceAccessor:
        return ResourceAccessor.existing_instance(username = username, resource_uri = resource_uri)



import re
import secrets
import string
import subprocess
from getpass import getpass

import keyring
import requests
from core_10x.global_cache import cache
from core_10x.rc import RC

from infra_10x.mongodb_admin import MongodbAdmin
from infra_10x.mongodb_store import MongoStore
from infra_10x.namespace import AUTO_PASSWORD_LENGTH, USER_ROLE


@cache
def create_xx_role(admin: MongodbAdmin):
    admin.db.command(
        'updateRole' if admin.role_exists('xxUser') else 'createRole',
        'xxUser',
        privileges=[
            {
                'resource': {'anyResource': True},
                'actions': USER_ROLE.WORKER.value,
            }
        ],
        roles=[],
    )


def create_xx_user(admin: MongodbAdmin, xx_user: str, new_password: str) -> None:
    password = keyring.get_password('xxuser', xx_user) or new_password
    admin.update_user(xx_user, password, 'xxUser', keep_current_roles=False)
    if password == new_password:
        keyring.set_password('xxuser', xx_user, password)


def create_xx_admin(admin: MongodbAdmin, xx_admin: str, new_password: str) -> None:
    if xx_admin == admin.username:
        raise ValueError('cannot create self!')
    password = keyring.get_password('xxadmin', xx_admin) or new_password
    admin.update_user(xx_admin, password, 'userAdminAnyDatabase', 'readWriteAnyDatabase')
    if password == new_password:
        keyring.set_password('xxadmin', xx_admin, password)


def get_xx_admin(user: str):
    return f'{user}-admin'


@cache
def git_user():
    return requests.get('https://api.github.com/user', headers={'Authorization': f'token {git_token()}'}).json()['login'] or input('Username:')


@cache
def git_token():
    return subprocess.getoutput("git credential fill <<< 'protocol=https\nhost=github.com\n' | grep password | cut -d= -f2") or input('Token:')


@cache
def all_git_users():
    git_org = re.search(r'github\.com[/:]([\w\-\.]+)/', subprocess.check_output(['git', 'remote', 'get-url', 'origin'], text=True).strip()).group(1)
    return [
        user['login']
        for user in requests.get(f'https://api.github.com/orgs/{git_org}/members', headers={'Authorization': f'token {git_token()}'}).json()
    ]


def generate_password():
    return ''.join(secrets.choice(string.ascii_letters + string.digits + '!@#$%^&*') for _ in range(AUTO_PASSWORD_LENGTH))


def create_users(hostname):
    admin = MongodbAdmin(hostname=hostname, **get_credentials('xxadmin', hostname))
    create_xx_role(admin)
    for xx_user in all_git_users():
        xx_admin = get_xx_admin(xx_user)
        if xx_admin != admin.username:  # -- do not break current working account!
            create_xx_admin(admin, xx_admin, generate_password())
        create_xx_user(admin, xx_user, generate_password())


def clear_vault():
    for xx_user in all_git_users():
        if keyring.get_password('xxuser', xx_user):
            keyring.delete_password('xxuser', xx_user)
        if xx_user != git_user():  # -- do not break own admin account!
            xx_admin = get_xx_admin(xx_user)
            if keyring.get_password('xxadmin', xx_admin):
                keyring.delete_password('xxadmin', xx_admin)


def drop_users(hostname):
    admin = MongodbAdmin(hostname=hostname, **get_credentials('xxadmin', hostname))
    for xx_user in all_git_users():
        xx_admin = get_xx_admin(xx_user)
        if xx_admin != admin.username:  # -- do not break current working account!
            admin.delete_user(xx_admin)
        admin.delete_user(xx_user)


def get_credentials(service, hostname):
    if hostname == 'localhost':
        return dict(username='', password='')
    username = git_user()
    if service == 'xxadmin':
        username = get_xx_admin(username)
    password = keyring.get_password(service, username)
    if not password:
        password = getpass(f"{username}'s password for `{service}`:")
        keyring.set_password(service, username, password)
    return dict(username=username, password=password)


def copy_db(from_host: str, to_host: str, dbname: str, overwrite=False) -> RC:
    from_store = MongoStore.instance(hostname=from_host, dbname=dbname, **get_credentials('xxuser', from_host))
    to_store = MongoStore.instance(hostname=to_host, dbname=dbname, **get_credentials('xxuser', to_host))
    return from_store.copy_to(to_store, overwrite=overwrite)


if __name__ == '__main__':
    host = 'mongo10x.eastus2.cloudapp.azure.com'
    # drop_users(host)
    # clear_vault()
    # create_users(host)

    copy_db(from_host='localhost', to_host=host, dbname='mkt_data', overwrite=False).throw()



import asyncio

import rio.testing.browser_client
from ui_10x.rio.component_builder import DynamicComponent
from ui_10x.rio.widgets.button_group import ButtonGroup, RadioButton


async def test_button_group() -> None:
    group = ButtonGroup()
    btn1 = RadioButton(label='A', value='A')
    btn2 = RadioButton(label='B', value='B')
    group.add_button(btn1, 0)
    group.add_button(btn2, 1)

    async with rio.testing.BrowserClient(lambda: DynamicComponent(group)) as test_client:
        await asyncio.sleep(0.5)
        # Initially, nothing selected
        assert group.checked_id() == -1

        # Select first button
        btn1.set_checked(True)
        await test_client.wait_for_refresh()
        assert group.checked_id() == 0

        # Select second button
        btn2.set_checked(True)
        await test_client.wait_for_refresh()
        assert group.checked_id() == 1

        # Deselect second button
        btn2.set_checked(False)
        await test_client.wait_for_refresh()
        assert group.checked_id() == -1



import traceback

import pytest

from core_10x.rc import RC
from core_10x.trait_definition import T, RT
from core_10x.trait_method_error import TraitMethodError
from core_10x.traitable import Traitable, AnonymousTraitable


OUTPUTS = ['exception', 'object', 'args', 'value']
GROUPS = [
    {
        'serialize_nx': lambda x: x.serialize_object(False),
        's_deserialize': lambda x: x.__class__.deserialize_object(x.s_bclass,None,{'_id':'a','_rev':1,'s': 1000}),
    },
    {'y_get': lambda x: x.y},
    {'x_get': lambda x: x.x(0)},
    {'x_set': lambda x: x.set_value('x', 100, 10)},
]

bombing_methods = {'_'.join(OUTPUTS[: i + 1]): group for i, group in enumerate(GROUPS)}


@pytest.mark.parametrize(argnames=['cnt', 'key'], argvalues=enumerate(bombing_methods.keys()))
def test_trait_method_error(cnt, key):
    class X(Traitable):
        x: int
        y: int
        z: int = T(1000)
        s: int = T()
        t: AnonymousTraitable = T(T.EMBEDDED)

        @classmethod
        def bombing_method(cls, x):
            raise KeyError(x + 1)

        def x_get(self, x):
            return self.bombing_method(x)

        def y_get(self):
            return self.bombing_method(0)

        def x_set(self, trait, value, x) -> RC:
            return self.bombing_method(value)

        @classmethod
        def s_serialize(cls, value):
            return cls.bombing_method(value)

        @classmethod
        def s_deserialize(cls, trait, value):
            return cls.bombing_method(value)

    x = X()
    x.t = X()
    for m, f in bombing_methods[key].items():
        try:
            f(x)
        except TraitMethodError as e:
            e_str = str(e)
            if not cnt:
                if m == 's_deserialize':
                    assert 'value = 1000' in e_str
                    assert 'original exception = KeyError: 1001' in e_str
                else:
                    assert 'original exception = TypeError: test_trait_method_error.<locals>.X/' in e_str
                    assert " - embedded instance must be 'embeddable'" in e_str

            for i, output in enumerate(key.split('_')):
                expected = i < cnt + 1
                if cnt or (not expected and output != 'value'):
                    assert (f' {output} =' in e_str) is expected, f'did{" not" if expected else ""} find {output} for {key} in {e_str}'
                if cnt:
                    assert f"Failed in <class 'test_trait_method_error.test_trait_method_error.<locals>.X'>.{m.split('_')[0]}.{m}" in e_str

            tb_str = traceback.format_exc()
            if cnt:
                assert 'f(x)' in tb_str
                assert f"'{m}': lambda x: x." in tb_str
                assert 'bombing_method(' in tb_str
                assert 'raise KeyError(x + 1)' in tb_str
        else:
            assert False, "Expected Exception!" # noqa: B011



from __future__ import annotations

import rio
import ui_10x.platform_interface as i
from ui_10x.rio.component_builder import Widget


class TextEdit(Widget, i.TextEdit):
    s_component_class = rio.MultiLineTextInput

    def _make_kwargs(self, **kwargs):
        kw = super()._make_kwargs(**kwargs)
        handler = self.focus_out_event
        if handler.__func__ != TextEdit.focus_out_event:  # do not set unless implemented
            kw['on_lose_focus'] = self.callback(handler)
        # TODO: test clipboard interactions
        return kw

    def to_plain_text(self) -> str:
        return self['text']

    def set_plain_text(self, text: str):
        self['text'] = text

    def set_read_only(self, readonly: bool):
        self['is_sensitive'] = not readonly

    def focus_out_event(self, event): ...



from __future__ import annotations

import abc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
    from datetime import date


class ABCMetaSlotted(abc.ABCMeta):
    def __new__(cls, name, bases, class_dict, **kwargs):
        class_dict.setdefault('__slots__', ())
        return super().__new__(cls, name, bases, class_dict, **kwargs)


class Object(abc.ABC, metaclass=ABCMetaSlotted): ...


# fmt: off
QueuedConnection            = None
AutoConnection              = None
DirectConnection            = None
UniqueConnection            = None
BlockingQueuedConnection    = None
# fmt: on


class signal_decl(abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def __init__(self, *args): ...

    @abc.abstractmethod
    def connect(self, method, type=None): ...

    @abc.abstractmethod
    def emit(self, *args): ...


class MouseEvent(abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def is_left_button(self) -> bool: ...

    @abc.abstractmethod
    def is_right_button(self) -> bool: ...


class Point(abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def x(self) -> int: ...

    @abc.abstractmethod
    def y(self) -> int: ...


class SizePolicy(abc.ABC, metaclass=ABCMetaSlotted):
    MINIMUM_EXPANDING = None
    PREFERRED = None


class FontMetrics(abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def average_char_width(self) -> int: ...


class Color(abc.ABC, metaclass=ABCMetaSlotted): ...


class Widget(abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def __init__(self, *args, **kwargs) -> None: ...

    @abc.abstractmethod
    def set_layout(self, layout: Layout): ...

    @abc.abstractmethod
    def set_style_sheet(self, sh: str): ...

    @abc.abstractmethod
    def style_sheet(self) -> str: ...

    @abc.abstractmethod
    def set_enabled(self, enabled: bool): ...

    @abc.abstractmethod
    def set_geometry(self, *args): ...

    @abc.abstractmethod
    def width(self) -> int: ...

    @abc.abstractmethod
    def height(self) -> int: ...

    @abc.abstractmethod
    def map_to_global(self, point: Point) -> Point: ...

    # @abc.abstractmethod
    # def mouse_press_event(self, event: MouseEvent): ...
    #
    # @abc.abstractmethod
    # def focus_out_event(self, event): ...
    #
    @abc.abstractmethod
    def set_size_policy(self, *args): ...

    @abc.abstractmethod
    def set_minimum_width(self, width: int): ...

    # @abc.abstractmethod
    # def set_maximum_width(self, width: int): ...

    @abc.abstractmethod
    def set_minimum_height(self, height: int): ...

    @abc.abstractmethod
    def font_metrics(self) -> FontMetrics: ...


# fmt: off
Horizontal  = 0
Vertical    = 1
# fmt: on


class Layout(abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def __init__(self, *args, **kwargs): ...

    @abc.abstractmethod
    def add_widget(self, widget: Widget, **kwargs): ...

    @abc.abstractmethod
    def set_spacing(self, spacing: int): ...

    @abc.abstractmethod
    def set_contents_margins(self, left, top, right, bottom): ...


class BoxLayout(Layout, abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def add_layout(self, layout: Layout, **kwargs): ...


class HBoxLayout(BoxLayout, abc.ABC, metaclass=ABCMetaSlotted): ...


class VBoxLayout(BoxLayout, abc.ABC, metaclass=ABCMetaSlotted): ...


class FormLayout(Layout):
    @abc.abstractmethod
    def add_row(self, w1: Widget, w2: Widget = None): ...


class Splitter(Widget):
    @abc.abstractmethod
    def add_widget(self, widget: Widget): ...

    @abc.abstractmethod
    def set_handle_width(self, width: int): ...

    @abc.abstractmethod
    def set_stretch_factor(self, widget_index: int, factor: int): ...

    @abc.abstractmethod
    def replace_widget(self, widget_index: int, widget: Widget): ...


class Label(Widget):
    @abc.abstractmethod
    def set_text(self, text: str): ...


class Style(abc.ABC, metaclass=ABCMetaSlotted):
    State_Active = None

    @abc.abstractmethod
    def standard_icon(self, style_icon): ...


class PushButton(Label):
    @abc.abstractmethod
    def clicked_connect(self, bound_method): ...

    @abc.abstractmethod
    def set_flat(self, flat: bool): ...


class LineEdit(Label, abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def text(self) -> str: ...

    @abc.abstractmethod
    def text_edited_connect(self, bound_method): ...

    @abc.abstractmethod
    def editing_finished_connect(self, bound_method): ...

    @abc.abstractmethod
    def set_read_only(self, readonly: bool): ...

    @abc.abstractmethod
    def set_password_mode(self): ...

    @abc.abstractmethod
    def set_tool_tip(self, tooltip: str): ...

    @abc.abstractmethod
    def mouse_press_event(self, event: MouseEvent): ...


class TextEdit(Widget, abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def to_plain_text(self) -> str: ...

    @abc.abstractmethod
    def set_plain_text(self, text: str): ...

    @abc.abstractmethod
    def set_read_only(self, readonly: bool): ...

    @abc.abstractmethod
    def focus_out_event(self, event): ...


class CheckBox(Label):
    @abc.abstractmethod
    def set_checked(self, checked: bool): ...

    @abc.abstractmethod
    def is_checked(self) -> bool: ...

    @abc.abstractmethod
    def state_changed_connect(self, bound_method): ...


class ComboBox(Widget): ...


class GroupBox(Widget):
    @abc.abstractmethod
    def set_title(self, title: str): ...


class RadioButton(Widget):
    @abc.abstractmethod
    def set_checked(self, checked: bool): ...


class ButtonGroup(Widget):
    @abc.abstractmethod
    def add_button(self, rb: RadioButton, id: int): ...

    @abc.abstractmethod
    def button(self, id: int) -> RadioButton: ...

    @abc.abstractmethod
    def checked_id(self) -> int: ...


class ListItem(abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def row(self): ...

    @abc.abstractmethod
    def set_selected(self, selected: bool): ...


MatchExactly = None


class ListWidget(Widget):
    @abc.abstractmethod
    def add_item(self, item): ...

    @abc.abstractmethod
    def add_items(self, items: list): ...

    @abc.abstractmethod
    def clicked_connect(self, bound_method): ...

    @abc.abstractmethod
    def clear(self): ...

    @abc.abstractmethod
    def find_items(self, text, flags): ...

    @abc.abstractmethod
    def row(self, item: ListItem) -> int: ...

    @abc.abstractmethod
    def take_item(self, row: int): ...


class TreeItem(Widget):
    @abc.abstractmethod
    def child_count(self): ...

    @abc.abstractmethod
    def set_expanded(self, expanded: bool): ...

    @abc.abstractmethod
    def set_text(self, col: int, text: str): ...

    @abc.abstractmethod
    def set_tool_tip(self, col: int, tooltip: str): ...


class TreeWidget(Widget):
    @abc.abstractmethod
    def set_column_count(self, column_count: int): ...

    @abc.abstractmethod
    def set_header_labels(self, labels: list): ...

    @abc.abstractmethod
    def top_level_item_count(self) -> int: ...

    @abc.abstractmethod
    def top_level_item(self, i: int) -> TreeItem: ...

    @abc.abstractmethod
    def resize_column_to_contents(self, col: int): ...

    @abc.abstractmethod
    def item_expanded_connect(self, item): ...

    @abc.abstractmethod
    def item_clicked_connect(self, bound_method): ...

    @abc.abstractmethod
    def item_pressed_connect(self, bound_method): ...

    @abc.abstractmethod
    def item_changed_connect(self, bound_method): ...

    @abc.abstractmethod
    def edit_item(self, item, col: int): ...

    @abc.abstractmethod
    def open_persistent_editor(self, item, col: int): ...


class CalendarWidget(Widget):
    @abc.abstractmethod
    def set_grid_visible(self, grid_visible: bool): ...

    @abc.abstractmethod
    def set_selected_date(self, selected_date: date): ...

    @abc.abstractmethod
    def selected_date(self) -> date: ...


class Dialog(Widget):
    @abc.abstractmethod
    def set_window_title(self, title: str): ...

    @abc.abstractmethod
    def set_window_flags(self, flags): ...

    @abc.abstractmethod
    def reject(self): ...

    @abc.abstractmethod
    def done(self, rc: int): ...

    @abc.abstractmethod
    def set_modal(self, modal: bool): ...

    @abc.abstractmethod
    def exec(self): ...

    @abc.abstractmethod
    def show(self): ...


class MessageBox(abc.ABC, metaclass=ABCMetaSlotted):
    @classmethod
    @abc.abstractmethod
    def question(cls, parent: Widget, title: str, message: str, on_close: Callable[[Dialog], None]) -> bool:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def warning(cls, parent: Widget, title: str, message: str, on_close: Callable[[Dialog], None]):
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def information(cls, parent: Widget, title: str, message: str, on_close: Callable[[Dialog], None]):
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def is_yes_button(cls, sb) -> bool:
        raise NotImplementedError


class Application(abc.ABC, metaclass=ABCMetaSlotted):
    @classmethod
    @abc.abstractmethod
    def instance(cls) -> Application: ...

    @abc.abstractmethod
    def style(self): ...


class TEXT_ALIGN:
    # fmt: off
    TOP       = None
    V_CENTER  = None
    BOTTOM    = None
    LEFT      = None
    CENTER    = None
    RIGHT     = None
    # fmt: on


class SCROLL(abc.ABC, metaclass=ABCMetaSlotted):
    # fmt: off
    OFF         = None
    ON          = None
    AS_NEEDED   = None
    # fmt: on


class ScrollArea(Widget):
    @abc.abstractmethod
    def set_widget(self, w: Widget): ...

    @abc.abstractmethod
    def set_horizontal_scroll_bar_policy(self, h): ...

    @abc.abstractmethod
    def set_vertical_scroll_bar_policy(self, h): ...


class StandardItem(abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def __init__(self, *args): ...

    @abc.abstractmethod
    def append_column(self, column: tuple): ...


class StandardItemModel(abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def set_item(self, row: int, col: int, item: StandardItem): ...

    @abc.abstractmethod
    def column_count(self, *args) -> int: ...

    @abc.abstractmethod
    def index(self, row: int, col: int): ...


class ModelIndex(abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def is_valid(self) -> bool: ...

    @abc.abstractmethod
    def child(self, row: int, col: int): ...

    @abc.abstractmethod
    def parent(self): ...

    @abc.abstractmethod
    def model(self) -> StandardItemModel: ...

    @abc.abstractmethod
    def data(self, role): ...


class AbstractTableModel(abc.ABC, metaclass=ABCMetaSlotted): ...


class HeaderView(abc.ABC, metaclass=ABCMetaSlotted):
    @abc.abstractmethod
    def set_model(self, model): ...

    @abc.abstractmethod
    def section_resized_connect(self, bound_method): ...

    @abc.abstractmethod
    def init_style_option(self, *args): ...


class Palette(abc.ABC, metaclass=ABCMetaSlotted):
    # fmt: off
    ButtonText  = None
    Button      = None
    Window      = None
    # fmt: on

    @abc.abstractmethod
    def set_brush(self, br_type, brush): ...


class StyleOptionHeader(abc.ABC, metaclass=ABCMetaSlotted):
    palette: Palette


ForegroundRole = None
BackgroundRole = None


class Separator(abc.ABC, metaclass=ABCMetaSlotted): ...


# fmt: off
def init(style = '') -> Application:           raise NotImplementedError
def to_clipboard(text: str, **kwargs):         raise NotImplementedError
def from_clipboard(**kwargs) -> str:           raise NotImplementedError
def separator(horizontal = True) -> Separator: raise NotImplementedError
def is_ui_thread() -> bool:                    raise NotImplementedError
# fmt: on



if __name__ == '__main__':

       from xx_common.rdate import RDate


       r = RDate('3M')
       r2 = RDate('10Y')

       c = r.conversion_freq_multiplier(r2.freq)

       x = r.equate_freq(r2)




from __future__ import annotations

import asyncio
import pathlib
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING

import ordered_set
from webview_proc import WebViewProcess

import rio
from rio import app_server, errors, utils

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

    import uvicorn


@dataclass
class App10x:
    """
    # App10x - a custom App to use in dialog.
    # rio.App is t.final, so App10x wraps rio.App and implements run_in_window
    # --> run webview out-of-process
    # --> run fastapi server in-process
    # --> uses a subclass fo FastAPIServer to create a custom session class
    #   --> custom session class communicates with out-of-process webview (rather than using webview_shim)
    """

    app: rio.App
    webview: WebViewProcess | None = None

    @staticmethod
    def _update_window_size(width: float | None, height: float | None) -> None:
        import webview  # imported here as called in separate process

        if width is None and height is None:
            return
        window = webview.windows[0]
        print(f'Current window size: {window.width}x{window.height} pixels')
        pixels_per_rem = window.evaluate_js("""
            let measure = document.createElement('div');
            document.body.appendChild(measure);
            measure.style.height = '1rem';
            let pixels_per_rem = measure.getBoundingClientRect().height * window.devicePixelRatio;
            measure.remove();
            pixels_per_rem;
        """)
        width_in_pixels = window.width if width is None else round(width * pixels_per_rem)
        height_in_pixels = window.height if height is None else round(height * pixels_per_rem)
        print(f'Resizing window to {width_in_pixels}x{height_in_pixels} pixels ({width}x{height} rem at {pixels_per_rem} pixels/rem)')
        window.resize(width_in_pixels, height_in_pixels)

    def _run_in_window(
        self,
        *,
        quiet: bool = True,
        maximized: bool = False,
        fullscreen: bool = False,
        width: float | None = None,
        height: float | None = None,
        debug_mode: bool = False,
        on_server_created: Callable[[uvicorn.Server], None] | None = None,
    ) -> None:
        host = 'localhost'
        port = utils.ensure_valid_port(host, None)
        url = f'http://{host}:{port}'

        server: uvicorn.Server | None = None

        icon_path = asyncio.run(self.app._fetch_icon_as_png_path())

        self.webview = webview = WebViewProcess(
            url=url,
            title=self.app.name,
            maximized=maximized,
            fullscreen=fullscreen,
            icon_path=icon_path,
            func=partial(self._update_window_size, width, height),
            on_close=lambda: server and setattr(server, 'should_exit', True),
        )

        def _on_server_created(_server: uvicorn.Server) -> None:
            nonlocal server
            server = _server

            fastapi_server = server.config.app
            fastapi_server.__class__ = FastapiServer
            fastapi_server.app10x = self
            if on_server_created:
                on_server_created(server)

        try:
            self.app._run_as_web_server(
                host=host,
                port=port,
                quiet=quiet,
                running_in_window=True,
                internal_on_app_start=webview.start,
                internal_on_server_created=_on_server_created,
                debug_mode=debug_mode,
            )
        except Exception as e:
            print(f'Error running app: {e}')
        finally:
            if webview.is_alive():
                webview.close()
                webview.join()


class FastapiServer(app_server.FastapiServer):
    app10x: App10x

    async def create_session(self, *args, **kwargs) -> rio.Session:
        session = await super().create_session(*args, **kwargs)
        session.__class__ = Session
        session.app10x = self.app10x
        return session


class Session(rio.Session):
    app10x: App10x

    async def _close(self, close_remote_session: bool) -> None:
        if not self.running_in_window:
            await super()._close(close_remote_session=close_remote_session)

        await super()._close(close_remote_session=False)
        if close_remote_session:
            self.app10x.webview.close()

    async def _get_webview_window(self):
        raise RuntimeError('Should not be called required in out-of-process webview')

    async def set_title(self, title: str) -> None:
        if not self.running_in_window:
            await super().set_title(title)

        self.app10x.webview.set_title(title)

    async def pick_folder(self) -> pathlib.Path:
        if not self.running_in_window:
            return await super().pick_folder()

        return pathlib.Path(self.app10x.webview.pick_folder())

    async def pick_file(
        self,
        *,
        file_types: Iterable[str] | None = None,
        multiple: bool = False,
    ) -> utils.FileInfo | list[utils.FileInfo]:
        if not self.running_in_window:
            return await super().pick_file(file_types=file_types, multiple=multiple)

        # Normalize the file types
        if file_types is not None:
            # Normalize and deduplicate, but maintain the order
            file_types = list(ordered_set.OrderedSet(utils.normalize_file_extension(file_type) for file_type in file_types))

        selected = self.app10x.webview.pick_file(
            file_types=[f'{extension} (*.{extension})' for extension in file_types],
            multiple=multiple,
        )

        if not selected:
            raise errors.NoFileSelectedError()

        return [utils.FileInfo._from_path(path) for path in selected] if multiple else utils.FileInfo._from_path(selected)

    async def save_file(
        self,
        file_contents: pathlib.Path | str | bytes,
        file_name: str = 'Unnamed File',
        *,
        media_type: str | None = None,
        directory: pathlib.Path | None = None,
    ) -> None:
        if not self.running_in_window:
            return await super().save_file(file_contents, file_name, media_type=media_type, directory=directory)

        self.app10x.webview.save_file(
            file_contents=file_contents,
            directory='' if directory is None else str(directory),
            file_name=file_name,
        )


Session._local_methods_ = rio.Session._local_methods_.copy()
Session._remote_methods_ = rio.Session._remote_methods_.copy()



from __future__ import annotations

import inspect
from collections import deque
from typing import TYPE_CHECKING

from core_10x.global_cache import singleton
from core_10x.named_constant import NamedConstant
from core_10x.rc import RC

from ui_10x.platform import ux

if TYPE_CHECKING:
    from collections.abc import Callable
    from datetime import date

    from core_10x.directory import Directory


class UxAsync(ux.Object):
    SIGNAL = ux.signal_decl()
    s_instances = {}

    @classmethod
    def init(cls, bound_method):
        instance = cls.s_instances.get(bound_method)
        if not instance:
            if not ux.is_ui_thread():
                raise RuntimeError(f'You must have called UxAsync.init({bound_method.__name__}) in the UI thread')

            instance = cls()
            cls.s_instances[bound_method] = instance

            instance.SIGNAL.connect(bound_method)

        return instance

    @classmethod
    def call(cls, bound_method):
        instance = cls.init(bound_method)
        instance.SIGNAL.emit()


class UxRadioBox(ux.GroupBox):
    __slots__ = ('group', 'nm_class', 'values',)
    def __init__(self, named_constant_class, title = '', horizontal = False, default_value: NamedConstant = None):
        assert inspect.isclass(named_constant_class) and issubclass(named_constant_class, NamedConstant), 'subclass of NamedConstant is expected'

        self.nm_class = named_constant_class
        self.values = values = list(named_constant_class.s_dir.values())

        if default_value is None:
            default = 0
        else:
            assert isinstance(default_value, named_constant_class), f'{named_constant_class} - unknown default_values = {default_value}'
            for i, c_def in enumerate(values):
                if c_def is default_value:
                    default = i
                    break
            else:
                raise RuntimeError(f'{default_value} is not found')

        if title:
            super().__init__(title)
        else:
            super().__init__()


        lay = ux.HBoxLayout() if horizontal else ux.VBoxLayout()
        self.set_layout(lay)

        self.group = group = ux.ButtonGroup()
        for i, c_def in enumerate(named_constant_class.s_dir.values()):
            rb = ux.RadioButton(c_def.label)
            lay.add_widget(rb)
            group.add_button(rb, id = i)
        group.button(default).set_checked(True)

    def choice(self):
        i = self.group.checked_id()
        return self.values[i]

def ux_success(text: str, parent = None, title = '', on_close: Callable[[bool],None] = None):
    if not title:
        title = 'Success'
    if not parent:
        parent = None
    ux.MessageBox.information(parent, title, text, on_close=on_close)

def ux_warning(text: str, parent = None, title = '',  on_close: Callable[[bool],None] = None):
    if not title:
        title = 'Warning'
    if not parent:
        parent = None
    ux.MessageBox.warning(parent, title, text, on_close=on_close)

def ux_answer(question: str, parent = None, title = '', on_close: Callable[[bool],None] = None) -> bool|None:
    if not title:
        title = 'Waiting for your answer...'
    if not parent:
        parent = None
    sb = ux.MessageBox.question(parent, title, question, on_close=(lambda result: on_close(ux.MessageBox.is_yes_button(result))) if on_close else None)
    if not on_close:
        print(f'ux_answer: question = {question}, title = {title}, parent = {parent}', sb)
        return ux.MessageBox.is_yes_button(sb)
    return None


def ux_multi_choice_answer(named_constant_class, parent = None, title = '', default_value: NamedConstant = None,  on_close: Callable[[NamedConstant],None] = None):
    if not title:
        title = 'Pick one of the choices below:'

    if not parent:
        parent = None

    box = UxRadioBox(named_constant_class, default_value = default_value)
    accept_callback = (lambda: on_close(box.choice())) if on_close else None
    d = UxDialog(box, parent = parent, title = title, cancel = '', accept_callback=accept_callback)
    if not on_close:
        d.exec()
        return box.choice()
    d.show()

def ux_push_button(label: str, callback = None, style_icon = None, flat = False):
    if style_icon:
        if isinstance(style_icon, str):
            try:
                style_icon = getattr(ux.Style.StandardPixmap, f'SP_{style_icon}')
            except AttributeError as e:
                raise RuntimeError(f"Unknown style_icon = '{style_icon}'") from e

        assert isinstance(style_icon, int), 'Currently only str or int are supported for style_icon'

        style = ux.Application.style()
        icon = style.standard_icon(style_icon)
        button = ux.PushButton(icon, label)
    else:
        button = ux.PushButton(label)

    if callback:
        button.clicked_connect(callback)

    if flat:
        button.set_flat(True)

    return button

class UxDialog(ux.Dialog):
    __slots__ = ('accept_callback','cancel_callback','w_message',)
    def _create_button(self, ok: bool, button_spec) -> ux.PushButton|None:
        if not button_spec:
            return None

        if isinstance(button_spec, str):
            label = button_spec
            icon = ''

        elif isinstance(button_spec, tuple):
            try:
                label, icon = button_spec
            except Exception:
                return None

        else:
            return None

        cb = self.on_ok if ok else self.on_cancel
        return ux_push_button(label, callback = cb, style_icon = icon)

    def __init__(
            self,
            w: ux.Widget,
            parent: ux.Widget   = None,
            title: str          = None,
            ok                  = 'Ok',
            cancel              = 'Cancel',
            accept_callback     = None,
            cancel_callback     = None,
            min_width           = 0,
            min_height          = 0,
            #window_flags        = None
    ):
        """
        :param w:               a Widget to show
        :param parent:          parent widget, if any
        :param title:           dialog title
        :param accept_callback: context.method, where: method( context ) -> RC
        :param ok, cancel:      labels and corresponding roles for Ok and Cancel buttons. No button if empty
        """
        super().__init__(parent)
        if title:
            self.set_window_title( title )

        # if window_flags is None:
        #     window_flags = Qt.WindowType.CustomizeWindowHint | Qt.WindowType.Window | Qt.WindowType.WindowTitleHint | Qt.WindowType.WindowCloseButtonHint

        ##self.setWindowFlags(window_flags)
        self.accept_callback = accept_callback
        self.cancel_callback = cancel_callback if cancel_callback else self.reject

        if ok:
            ok = self._create_button(True, ok)
        if cancel:
            cancel = self._create_button(False, cancel)

        lay = ux.VBoxLayout()
        self.set_layout(lay)

        lay.add_widget(w)

        self.w_message = ux.Label()
        self.w_message.set_style_sheet('color: red;')
        lay.add_widget(self.w_message)

        if ok or cancel:
            lay.add_widget(ux.separator())

            bar = ux.HBoxLayout()
            lay.add_layout(bar)

            bar.add_widget(ux.Label(), stretch = 1)
            if ok:
                bar.add_widget(ok)
            if cancel:
                bar.add_widget(cancel)

        if min_width > 0:
            self.set_minimum_width(min_width)
        if min_height > 0:
            self.set_minimum_height(min_height)

    def on_ok(self):
        if self.accept_callback:
            rc: RC = self.accept_callback()
            if not rc:
                self.message(rc.error())
                return

        self.done(1)

    def on_cancel(self):
        self.reject()
        self.done(0)

    def message(self, text: str):
        self.w_message.set_text(text)

def ux_pick_date(title = 'Pick a Date', show_date: date = None, grid = True, default = None, on_accept = None) -> date|None:
    cal = ux.CalendarWidget()
    cal.set_grid_visible(bool(grid))
    if show_date:
        cal.set_selected_date(show_date)

    accept_callback = (lambda: on_accept(cal.selected_date())) if on_accept else None
    dlg = UxDialog(cal, title = title, accept_callback = accept_callback)
    if on_accept:
        dlg.show()
        return None
    return cal.selected_date() if dlg.exec() else default

class UxStyleSheet:
    def __init__(self, widget):
        self.widget = widget
        sh = widget.style_sheet()
        self.data = self.loads(sh)
        self.replacement_stack = deque()

    def set(self):
        self.widget.set_style_sheet(self.dumps(self.data))

    def update(self, named_sheet_attrs: dict, _system = False):
        if not named_sheet_attrs:
            return

        data = self.data
        if not any(value != data.get(name) for name, value in named_sheet_attrs.items()):   #-- no attribute values changed, nothing to do!
            return

        if _system:
            old_data = { name: data.get(name) for name in named_sheet_attrs.keys() }
            self.replacement_stack.append(old_data)

        data.update(named_sheet_attrs)
        self.set()

    def restore(self):
        if self.replacement_stack:
            old_data = self.replacement_stack.pop()
            data = self.data
            for name, value in old_data.items():
                if value is None:
                    data.pop(name, None)
                else:
                    data[name] = value

            self.set()

    @classmethod
    def dumps(cls, data: dict) -> str:
        return '\n'.join(f'{name}: {value};' for name, value in data.items())

    @classmethod
    def loads(cls, sheet: str) -> dict:
        res = {}
        pairs = sheet.split(';')
        for pair in pairs:
            pair = pair.strip()
            if pair:
                name_value = pair.split(':')
                if len(name_value) == 2:
                    name = name_value[0].strip()
                    value = name_value[1].strip()
                    res[name] = value

        return res

    @classmethod
    def modify(cls, widget, named_sheet_attrs: dict):
        if not named_sheet_attrs:
            return

        sh = widget.style_sheet()
        data = cls.loads(sh)
        data.update(named_sheet_attrs)
        widget.set_style_sheet(cls.dumps(data))

    @classmethod
    def _color(cls, style_sheet: str, attr_name: str) -> ux.Color:
        data = cls.loads(style_sheet)
        str_color = data.get(attr_name)
        return ux.Color(str_color) if str_color else None

    @classmethod
    def fg_color(cls, style_sheet: str) -> ux.Color:
        return cls._color(style_sheet, 'color')

    @classmethod
    def bg_color(cls, style_sheet: str) -> ux.Color:
        return cls._color(style_sheet, 'background-color')

    @classmethod
    def color(cls, style_sheet: str, background = False) -> ux.Color:
        return cls.bg_color(style_sheet) if background else cls.fg_color(style_sheet)


s_verticalAlignmentMap = { # noqa: N816
    -1:     ux.TEXT_ALIGN.TOP,
    0:      ux.TEXT_ALIGN.V_CENTER,
    1:      ux.TEXT_ALIGN.BOTTOM,
}
s_horizontalAlignmentMap = { # noqa: N816
    -1:     ux.TEXT_ALIGN.LEFT,
    0:      ux.TEXT_ALIGN.CENTER,
    1:      ux.TEXT_ALIGN.RIGHT,
}
def ux_text_alignment(align_value: int, horizontal = True) -> int:
    if horizontal:
        return s_horizontalAlignmentMap.get(align_value, ux.TEXT_ALIGN.RIGHT) | ux.TEXT_ALIGN.V_CENTER

    raise NotImplementedError

def ux_make_scrollable(w: ux.Widget, h = ux.SCROLL.AS_NEEDED, v = ux.SCROLL.AS_NEEDED) -> ux.Widget:
    if h == ux.SCROLL.OFF and v == ux.SCROLL.OFF:
        return w

    sa = ux.ScrollArea()
    sa.set_widget(w)
    sa.set_horizontal_scroll_bar_policy(h)
    sa.set_vertical_scroll_bar_policy(v)
    return sa

class UxSearchableList(ux.GroupBox):
    __slots__ = ('case_sensitive', 'current_choices', 'hook', 'initial_choices', 'reset_selection', 'selection', 'sort', 'w_field', 'w_list')
    def __init__(
        self,
        text_widget: ux.LineEdit = None,
        reset_selection = True,
        choices: list   = None,
        select_hook     = None,
        sort            = False,
        case_sensitive  = False,
        title           = ''
    ):
        """
        :param text_widget: if an external text_widget given, it is used, otherwise its own is created on top of the list
        :param reset_selection: if True, text_widget will be reset on new selection
        :param choices: all choices
        :param select_hook: if given, will be called after a new selection is made: select_hook( selected_choice: str )
        :param sort: sort choices if True
        :param case_sensitive: whether to search with case sensitivity
        :param title: an optional title
        """
        super().__init__()
        self.sort = sort
        self.initial_choices = choices if not sort else sorted(choices)
        self.current_choices = self.initial_choices
        self.hook = select_hook
        self.reset_selection = reset_selection
        self.case_sensitive = case_sensitive
        self.selection = None

        if title:
            self.set_title(title)
        lay = ux.VBoxLayout()

        if text_widget:
            self.w_field = text_widget
        else:
            self.w_field = ux.LineEdit()
            lay.add_widget(self.w_field)
        self.w_field.text_edited_connect(self.process_input)

        self.w_list = ux.ListWidget()

        self.w_list.add_items(self.initial_choices)
        self.w_list.clicked_connect(self.item_selected)
        lay.add_widget(self.w_list)

        self.set_layout(lay)

    def add_choice(self, choice: str):
        if choice not in self.initial_choices:
            self.initial_choices.append(choice)
            if self.sort:
                self.initial_choices = sorted(self.initial_choices)

            self.w_list.clear()
            self.w_list.add_items(self.initial_choices)

    def remove_choice(self, choice: str):
        if choice in self.initial_choices:
            self.initial_choices.remove(choice)
            self.w_list.clear()
            self.w_list.add_items(self.initial_choices)

    def process_input(self, input):
        self.w_list.clear()
        if not input:
            self.current_choices = self.initial_choices
        else:
            if not self.current_choices:
                self.current_choices = self.initial_choices

            if not self.case_sensitive:
                input = input.lower()
                self.current_choices = [s for s in self.current_choices if not s.lower().find(input) == -1]
            else:
                self.current_choices = [s for s in self.current_choices if not s.find(input) == -1]

        self.w_list.add_items(self.current_choices)

    def item_selected(self, item: ux.ListItem):
        i = item.row()
        text = self.current_choices[i]
        self.reset()

        items = self.w_list.find_items(text, ux.MatchExactly)
        if len(items):
            items[0].set_selected(True)
            text_s = '' if self.reset_selection else text
            self.w_field.set_text(text_s)

        self.selection = text
        if self.hook:
            self.hook(text)

    def choice(self) -> str:
        return self.selection

    def reset(self):
        if not self.current_choices == self.initial_choices:
            self.current_choices = self.initial_choices
            self.w_list.clear()
            self.w_list.add_items( self.initial_choices)

class UxTreeViewer(ux.TreeWidget):
    s_label_max_length  = 40

    def __init__(self, dir: Directory, select_hook = None, label_max_length = -1, expand = False, **kwargs):
        super().__init__()
        if label_max_length < 0:
            label_max_length = self.s_label_max_length

        self.dir = dir
        self.select_hook = select_hook

        dir_value = dir.value
        dir_name = dir.name
        if dir_value and dir_name and dir_name != dir_value:
            num_cols = 2
            header_labels = [dir_name, 'Description']
        else:
            num_cols = 1
            header_labels = [dir.show_value()]

        self.num_cols = num_cols
        self.set_column_count(num_cols)
        self.set_header_labels(header_labels)
        self.item_clicked_connect(self.tree_item_clicked)

        for subdir in dir.members.values():     #-- without the root!
            self.create_tree(subdir, self, label_max_length, num_cols)

        if expand:
            for i in range(self.top_level_item_count()):
                top = self.top_level_item(i)
                top.set_expanded(True)

        self.resize_column_to_contents(0)
        if num_cols == 2:
            self.resize_column_to_contents(1)
        self.set_size_policy(ux.SizePolicy.MINIMUM_EXPANDING, ux.SizePolicy.MINIMUM_EXPANDING)

    class TreeItem(ux.TreeItem):
        def __init__(self, parent_node, dir: Directory):
            super().__init__(parent_node)
            self.dir = dir

    @classmethod
    def create_tree(cls, dir: Directory, parent_node, label_max_length: int, num_cols: int):
        node = cls.TreeItem(parent_node, dir)
        show_value = dir.show_value()
        node.set_text(0, show_value)
        if num_cols == 2:
            label = dir.name
            if label and label != show_value:
                if len(label) > label_max_length:
                    label = label[ :label_max_length ] + '...'

                node.set_text(1, label)

        node.set_tool_tip(num_cols-1, dir.name)

        for subdir in dir.members.values():
            cls.create_tree(subdir, node, label_max_length, num_cols)

    def tree_item_clicked(self, item: TreeItem):
        self.on_tag_selected(item.dir)

    def on_tag_selected(self, dir: Directory):
        if self.select_hook:
            self.select_hook(dir)

# class UxPixmap:
#     COLOR_AUTO      = Qt.ColorScheme.AutoColor
#     COLOR_DITHER    = Qt.ColorScheme.ColorOnly
#     COLOR_MONO      = Qt.ColorScheme.MonoOnly
#     RATIO_IGNORE    = Qt.ColorScheme.IgnoreAspectRatio
#     RATIO_KEEP      = Qt.ColorScheme.KeepAspectRatio
#     RATIO_KEEP_EXP  = Qt.ColorScheme.KeepAspectRatioByExpanding
#
#     s_sourceTypeMap = {
#         str:        QPixmap.load,
#         bytes:      QPixmap.loadFromData,
#         bytearray:  QPixmap.loadFromData,
#     }
#     def __init__( self, source = None, image_format = '', color_flags = COLOR_AUTO ):
#         self.m_pixmap = QPixmap()
#         rc = self.load( source, image_format = image_format, color_flags = color_flags )
#         if not rc:
#             raise RuntimeError( rc.err() )
#
#     def load( self, source, image_format = '', color_flags = COLOR_AUTO ) -> RC:
#         if not source:
#             return RC( True )
#
#         method = self.s_sourceTypeMap.get( type( source ) )
#         if not method:
#             return RC( False, f'UxPixmap - unknown type of source = {type( source )}' )
#
#         rc = method( self.m_pixmap, source, format = image_format, flags = color_flags )
#         if not rc:
#             return RC( False, f'UxPixmap - failed to initialize from source = {source}' )
#
#         return RC( True )
#
#     def _toData( self, array: bytearray, format = '', quality = -1 ) -> bool:
#         buffer = QBuffer( array )
#         buffer.open( QIODevice.WriteOnly )
#         return self.m_pixmap.save( buffer, format = format, quality = quality )
#
#     s_destinationTypeMap = {
#         str:        QPixmap.save,
#         bytearray:  _toData,
#     }
#     def save( self, destination, color_format = '', quality = -1 ) -> RC:
#         method = self.s_destinationTypeMap.get( type( destination ) )
#         if not method:
#             return RC( False, f'UxPixmap - unknown destination type = {type( destination )}' )
#
#         rc = method( self, destination, format = color_format, quality = quality )
#         if not rc:
#             return RC( False, f'UxPixmap - failed to save to destination = {destination}' )
#
#         return RC( True )
#
#     def render( self, label: QLabel, w = 0, h = 0, scaled = RATIO_IGNORE, smooth = True ):
#         transform_mode = Qt.SmoothTransformation if smooth else Qt.FastTransformation
#         pixmap = self.m_pixmap
#         if w > 0 and h > 0:
#             pixmap = pixmap.scaled( w, h, aspectRatioMode = scaled, transformMode = transform_mode )
#         label.setPixmap( pixmap )

@singleton
class UxClipBoard:
    def __init__(self):
        self.dir = {}
        self.entity = None
        self.trait_name = ''

    # def copy(self, tag: str, value):
    #     self.dir[tag] = value
    #
    # def paste(self, tag: str):
    #     return self.dir.get(tag)
    #
    # def clear(self):
    #     self.dir = {}
    #
    # def default_tag( self ) -> str:
    #     assert self.entity and self.trait_name, 'entity and trait name are not defined'
    #
    #     cls = self.entity.__class__
    #     trait = cls.trait(self.trait_name)
    #     assert trait, f"{cls}: unknown trait '{self.trait_name}'"
    #     #tag_label = trait.label if trait.label else self.trait_name
    #     tag_label = self.trait_name
    #     return f"{self.entity}: '{tag_label}'"
    #
    # def layout(self) -> QWidget:
    #     w = QWidget()
    #     lay = QVBoxLayout()
    #     w.setLayout(lay)
    #
    #     add_row = QHBoxLayout()
    #     add_b = uxPushButton('Add', callback = self.onCopy)
    #     self.m_addNameWidget = add_name = QLineEdit()
    #     add_name.setText(self.default_tag())
    #     add_row.addWidget(add_b)
    #     add_row.addWidget(add_name)
    #     lay.addLayout(add_row)
    #
    #     lay.addWidget(uxSeparator())
    #
    #     self.m_sl = sl = UxSearchableList(choices = list( self.dir.keys() ), select_hook = self.onPaste, title = 'Use Data')
    #     lay.addWidget(sl)
    #
    #     return w
    #
    # def popup(self, parent_widget: QWidget, entity, trait_name: str):
    #     self.m_parent = parent_widget
    #     self.entity = entity
    #     self.trait_name = trait_name
    #     self.m_d = d = UxDialog(self.layout(), parent = parent_widget, cancel = '', title = 'Copy or Paste Data')
    #     d.exec()
    #
    # def onCopy(self):
    #     name = self.m_addNameWidget.text()
    #     if name:
    #         value = self.entity.get_value( self.trait_name )
    #         if value is not None:
    #             self.copy( name, value )
    #
    #     self.m_d.close()
    #
    # def onPaste( self, name: str ):
    #     value = self.paste( name )
    #     if value is not None:
    #         rc = self.m_entity.setValue( self.m_traitName, value )
    #         if not rc:
    #             uxWarning(
    #                 f"Data '{name}' is not suitable for '{self.m_entity.trait( self.m_traitName ).m_label}'",
    #                 parent = self.m_parent,
    #                 title = 'Invalid Data'
    #             )
    #
    #     self.m_d.close()





from py10x_kernel import BTraitableProcessor

from core_10x.code_samples.person import Person
from core_10x.trait_definition import T


class MarriedPerson(Person):
    spouse: Person = T()


if __name__ == '__main__':
    from core_10x.manual_tests.reference_serialization_bug import MarriedPerson

    with BTraitableProcessor.create_root():
        MarriedPerson(first_name='Ilya', last_name='Pevzner', spouse=MarriedPerson(first_name='Tatiana', last_name='Pevzner'), _replace=True).save(
            save_references=True
        )

    assert type(MarriedPerson(first_name='Ilya', last_name='Pevzner').spouse) is MarriedPerson



import matplotlib.colors
from ui_10x.rio.style_sheet import StyleSheet

import rio


def test_ss_to_ts():
    sheet = {
        'color': 'lightgreen',
        'background-color': 'white',
        'font-family': 'Helvetica',
        'font-style': 'italic',
        'font-weight': 'bold',
        'border-width': '2px',
        'border-style': '',
        'border-color': 'blue',
    }
    ss = StyleSheet()
    rc = ss.set_values(sheet=sheet)
    assert not rc
    assert 'background-color' in rc.error()
    assert 'border-color' in rc.error()
    assert 'border-style' in rc.error()
    assert 'border-style' in rc.error()
    assert 'border-width' in rc.error()
    ts = ss.text_style
    assert ts.italic
    assert ts.font_weight == 'bold'
    assert ts.font._google_fonts_name == 'Helvetica'
    assert matplotlib.colors.same_color(f'#{ts.fill.hex}', 'lightgreen')


def test_ts_to_ss():
    ss = StyleSheet()
    rc = ss.set_values(text_style=rio.TextStyle(font=rio.Font.from_google_fonts('Times New Roman'), fill=rio.Color.from_hex('#c5f7c5')))
    assert rc
    assert ss.sheet == {'color': '#c5f7c5', 'font-family': 'Times New Roman', 'font-style': 'normal', 'font-weight': 'normal'}
